# Copyright (C) 2007, Eduardo Silva <edsiper@gmail.com>
# Copyright (C) 2008, One Laptop Per Child
# Copyright (C) 2009, Tomeu Vizoso
# Copyright (C) 2011, Benjamin Berg <benjamin@sipsolutions.net>
# Copyright (C) 2011, Marco Pesenti Gritti <marco@marcopg.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
STABLE.
"""
import textwrap

from gi.repository import GLib
from gi.repository import Gtk
from gi.repository import Gdk
from gi.repository import GObject
from gi.repository import Pango

from sugar3.graphics import style
from sugar3.graphics.icon import Icon
from sugar3.graphics.palettewindow import PaletteWindow, \
    _PaletteWindowWidget, _PaletteMenuWidget
from sugar3.graphics.palettemenu import PaletteMenuItem

from sugar3.graphics.palettewindow import MouseSpeedDetector, Invoker, \
    WidgetInvoker, CursorInvoker, ToolInvoker, TreeViewInvoker

assert MouseSpeedDetector
assert Invoker
assert WidgetInvoker
assert CursorInvoker
assert ToolInvoker
assert TreeViewInvoker


class _HeaderItem(Gtk.MenuItem):
    """A MenuItem with a custom child widget that gets all the
    available space.

    """

    __gtype_name__ = 'SugarPaletteHeader'

    def __init__(self, widget):
        Gtk.MenuItem.__init__(self)
        if self.get_child() is not None:
            self.remove(self.get_child())
        self.add(widget)
        # FIXME we have to mark it as insensitive again to make it an
        # informational element, when we realize how to get the icon
        # displayed correctly - SL #3836
        # self.set_sensitive(False)

    def do_size_allocate(self, allocation):
        self.set_allocation(allocation)
        self.get_child().size_allocate(allocation)

    def do_draw(self, cr):
        # force state normal, we don't want change color when hover
        self.set_state(Gtk.StateType.NORMAL)

        # draw separator
        allocation = self.get_allocation()
        cr.save()
        line_width = 2
        cr.set_line_width(line_width)
        cr.set_source_rgb(.5, .5, .5)  # button_grey #808080;
        cr.move_to(0, allocation.height - line_width)
        cr.line_to(allocation.width, allocation.height - line_width)
        cr.stroke()
        cr.restore()

        # draw content
        Gtk.MenuItem.do_draw(self, cr)
        return False


class Palette(PaletteWindow):
    """Floating palette implementation.

    This class dynamically switches between one of two encapsulated child
    widget types: a _PaletteWindowWidget or a _PaletteMenuWidget.

    The window widget, created by default, acts as the container for any
    type of widget the user may wish to add. It can optionally display primary
    text, secondary text, and an icon at the top of the palette.

    If the user attempts to access the 'menu' property, the window widget is
    destroyed and the palette is dynamically switched to use a menu widget.
    This is a GtkMenu that retains the same look and feel as a normal palette,
    allowing submenus and so on. If primary text, secondary text and/or icons
    were provided, an initial menu entry is created containing widgets to
    display such information.
    """

    __gsignals__ = {
        'activate': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    __gtype_name__ = 'SugarPalette'

    def __init__(self, label=None, accel_path=None,
                 text_maxlen=style.MENU_WIDTH_CHARS, **kwargs):
        # DEPRECATED: label is passed with the primary-text property,
        # accel_path is set via the invoker property

        self._primary_text = None
        self._secondary_text = None
        self._icon = None
        self._icon_visible = True

        self._primary_event_box = Gtk.EventBox()
        self._primary_event_box.show()
        self._primary_box = Gtk.HBox()
        self._primary_event_box.add(self._primary_box)
        self._primary_box.show()

        self._icon_box = Gtk.HBox()
        self._icon_box.set_size_request(style.GRID_CELL_SIZE, -1)
        self._primary_box.pack_start(self._icon_box, False, True, 0)

        labels_box = Gtk.VBox()
        self._label_alignment = Gtk.Alignment(xalign=0, yalign=0.5, xscale=1,
                                              yscale=0.33)
        self._label_alignment.set_padding(
            style.DEFAULT_SPACING, style.DEFAULT_SPACING,
            style.DEFAULT_SPACING, style.DEFAULT_SPACING)
        self._label_alignment.add(labels_box)
        self._label_alignment.show()
        self._primary_box.pack_start(self._label_alignment, True, True, 0)
        labels_box.show()

        self._label = Gtk.AccelLabel(label='')
        self._label.set_alignment(0, 0.5)

        if text_maxlen > 0:
            self._label.set_max_width_chars(text_maxlen)
            self._label.set_ellipsize(style.ELLIPSIZE_MODE_DEFAULT)
        labels_box.pack_start(self._label, True, True, 0)
        self._primary_event_box.connect('button-release-event',
                                        self.__button_release_event_cb)
        self._primary_event_box.set_events(Gdk.EventMask.BUTTON_RELEASE_MASK)

        self._secondary_label = Gtk.Label()
        self._secondary_label.set_alignment(0, 0.5)
        labels_box.pack_start(self._secondary_label, True, True, 0)

        self._secondary_box = Gtk.VBox()

        self._separator = Gtk.HSeparator()
        self._secondary_box.pack_start(self._separator, True, True, 0)
        self._secondary_box.show()

        # we init after initializing all of our containers
        PaletteWindow.__init__(self, **kwargs)

        self._full_request = [0, 0]
        self._content = None

        # we set these for backward compatibility
        if label is not None:
            self.props.primary_text = label

        self._add_content()

        self.action_bar = PaletteActionBar()
        self._secondary_box.pack_start(self.action_bar, True, True, 0)
        self.action_bar.show()

        self.connect('notify::invoker', self.__notify_invoker_cb)

        # Default to a normal window palette
        self._content_widget = None
        self.set_content(None)

    def _setup_widget(self):
        PaletteWindow._setup_widget(self)
        self._widget.connect('destroy', self.__destroy_cb)
        self._widget.connect('map', self.__map_cb)

    def __map_cb(self, *args):
        # Fixes #4463
        if hasattr(self._widget, 'present'):
            self._widget.present()

    def __destroy_cb(self, palette):
        self.popdown(immediate=True)
        # Break the reference cycle. It looks like the gc is not able to free
        # it, possibly because Gtk.Menu memory handling is very special.
        self._widget = None

    def __notify_invoker_cb(self, palette, pspec):
        invoker = self.props.invoker
        if invoker is not None and hasattr(invoker.props, 'widget'):
            self._update_accel_widget()
            self._invoker.connect('notify::widget',
                                  self.__invoker_widget_changed_cb)

    def __invoker_widget_changed_cb(self, invoker, spec):
        self._update_accel_widget()

    def get_full_size_request(self):
        return self._full_request

    def popup(self, immediate=False):
        if self._invoker is not None:
            self._update_full_request()

        PaletteWindow.popup(self, immediate)

    def popdown(self, immediate=False, state=None):
        '''
        Popdown (or show the full contents of) the palette.

        Keyword Args:
            immediate (bool): if True, the palette will be shown instantly (as
                if the user right clicked the item).  If False, the palette
                will be shown after the usual activation wait time.

        .. deprecated:: 0.109
            The state keyword argument is deprecated.  The old
            secondary state has become the only state.
        '''
        if immediate:
            if self._widget:
                self._widget.size_request()
        PaletteWindow.popdown(self, immediate)

    def on_enter(self):
        PaletteWindow.on_enter(self)

    def _add_content(self):
        # The content is not shown until a widget is added
        self._content = Gtk.VBox()
        self._secondary_box.pack_start(self._content, True, True,
                                       style.DEFAULT_SPACING)

    def _update_accel_widget(self):
        assert self.props.invoker is not None
        self._label.props.accel_widget = self.props.invoker.props.widget

    def set_primary_text(self, label, accel_path=None):
        self._primary_text = label
        if label is not None:
            label = GLib.markup_escape_text(label)
            self._label.set_markup('<b>%s</b>' % label)
            self._label.show()

    def get_primary_text(self):
        return self._primary_text

    primary_text = GObject.Property(type=str,
                                    getter=get_primary_text,
                                    setter=set_primary_text)

    def __button_release_event_cb(self, widget, event):
        if self.props.invoker is not None:
            self.props.invoker.primary_text_clicked()

    def set_secondary_text(self, label):
        if label is None:
            self._secondary_label.hide()
        else:
            NO_OF_LINES = 3
            ELLIPSIS_LENGTH = 6

            label = label.replace('\n', ' ')
            label = label.replace('\r', ' ')

            if hasattr(self._secondary_label, 'set_lines'):
                self._secondary_label.set_max_width_chars(
                    style.MENU_WIDTH_CHARS)
                self._secondary_label.set_line_wrap(True)
                self._secondary_label.set_ellipsize(
                    style.ELLIPSIZE_MODE_DEFAULT)
                self._secondary_label.set_line_wrap_mode(
                    Pango.WrapMode.WORD_CHAR)
                self._secondary_label.set_lines(NO_OF_LINES)
                self._secondary_label.set_justify(Gtk.Justification.FILL)
            else:
                # FIXME: fallback for Gtk < 3.10
                body_width = NO_OF_LINES * style.MENU_WIDTH_CHARS
                body_width -= ELLIPSIS_LENGTH
                if len(label) > body_width:
                    label = ' '.join(label[:body_width].split()[:-1]) + '...'
                label = textwrap.fill(label, width=style.MENU_WIDTH_CHARS)

            self._secondary_text = label
            self._secondary_label.set_text(label)
            self._secondary_label.show()

    def get_secondary_text(self):
        return self._secondary_text

    secondary_text = GObject.Property(type=str, getter=get_secondary_text,
                                      setter=set_secondary_text)

    def _show_icon(self):
        self._label_alignment.set_padding(
            style.DEFAULT_SPACING, style.DEFAULT_SPACING,
            0, style.DEFAULT_SPACING)
        self._icon_box.show()

    def _hide_icon(self):
        self._icon_box.hide()
        self._label_alignment.set_padding(
            style.DEFAULT_SPACING, style.DEFAULT_SPACING,
            style.DEFAULT_SPACING, style.DEFAULT_SPACING)

    def set_icon(self, icon):
        if icon is None:
            self._icon = None
            self._hide_icon()
        else:
            if self._icon:
                self._icon_box.remove(self._icon_box.get_children()[0])

            event_box = Gtk.EventBox()
            event_box.connect('button-release-event',
                              self.__icon_button_release_event_cb)
            self._icon_box.pack_start(event_box, True, True, 0)
            event_box.show()

            self._icon = icon
            self._icon.props.pixel_size = style.STANDARD_ICON_SIZE
            event_box.add(self._icon)
            self._icon.show()
            self._show_icon()

    def get_icon(self):
        return self._icon

    icon = GObject.Property(type=object, getter=get_icon, setter=set_icon)

    def __icon_button_release_event_cb(self, icon, event):
        self.emit('activate')

    def set_icon_visible(self, visible):
        self._icon_visible = visible

        if visible and self._icon is not None:
            self._show_icon()
        else:
            self._hide_icon()

    def get_icon_visible(self):
        return self._icon_visilbe

    icon_visible = GObject.Property(type=bool,
                                    default=True,
                                    getter=get_icon_visible,
                                    setter=set_icon_visible)

    def set_content(self, widget):
        assert self._widget is None \
            or isinstance(self._widget, _PaletteWindowWidget)

        if self._widget is None:
            self._widget = _PaletteWindowWidget(self)
            self._setup_widget()

            self._palette_box = Gtk.VBox()
            self._palette_box.pack_start(self._primary_event_box, False, True,
                                         0)
            self._palette_box.pack_start(self._secondary_box, True, True, 0)

            self._widget.add(self._palette_box)
            self._palette_box.show()
            height = style.GRID_CELL_SIZE - 2 * self._widget.get_border_width()
            self._primary_event_box.set_size_request(-1, height)

        if self._content.get_children():
            self._content.remove(self._content.get_children()[0])

        if widget is not None:
            widget.connect('button-release-event',
                           self.__widget_button_release_cb)
            self._content.add(widget)
            self._content.show()
        else:
            self._content.hide()

        self._content_widget = widget

        self._update_accept_focus()
        self._update_separators()

    def __widget_button_release_cb(self, widget, event):
        event_widget = Gtk.get_event_widget(event)
        if isinstance(event_widget, PaletteMenuItem):
            self.popdown(immediate=True)
        return False

    def get_label_width(self):
        # Gtk.AccelLabel request doesn't include the accelerator.
        label_width = self._label_alignment.get_preferred_width()[1] + \
            self._label.get_accel_width()
        return label_width

    def _update_separators(self):
        visible = self._content.get_children()
        self._separator.props.visible = visible

    def _update_accept_focus(self):
        accept_focus = len(self._content.get_children())
        self._widget.set_accept_focus(accept_focus)

    def _update_full_request(self):
        if self._widget is not None:
            self._full_request = self._widget.size_request()

    def get_menu(self):
        assert self._content_widget is None

        if self._widget is None \
                or not isinstance(self._widget, _PaletteMenuWidget):
            if self._widget is not None:
                self._palette_box.remove(self._primary_event_box)
                self._palette_box.remove(self._secondary_box)
                self._teardown_widget()
                self._widget.destroy()

            self._widget = _PaletteMenuWidget()

            self._label_menuitem = _HeaderItem(self._primary_event_box)
            self._label_menuitem.show()
            self._widget.append(self._label_menuitem)

            self._setup_widget()

        return self._widget

    menu = GObject.Property(type=object, getter=get_menu)

    def _invoker_right_click_cb(self, invoker):
        self.popup(immediate=True)


class PaletteActionBar(Gtk.HButtonBox):

    def add_action(self, label, icon_name=None):
        button = Gtk.Button(label)

        if icon_name:
            icon = Icon(icon_name)
            button.set_image(icon)
            icon.show()

        self.pack_start(button, True, True, 0)
        button.show()



# Copyright (C) 2007, Eduardo Silva <edsiper@gmail.com>
# Copyright (C) 2008, One Laptop Per Child
# Copyright (C) 2009, Tomeu Vizoso
# Copyright (C) 2011, Benjamin Berg <benjamin@sipsolutions.net>
# Copyright (C) 2011, Marco Pesenti Gritti <marco@marcopg.org>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
STABLE.
"""

import logging
import math

import gi
gi.require_version('SugarGestures', '1.0')

from gi.repository import Gdk
from gi.repository import Gtk
from gi.repository import GObject
from gi.repository import GLib

from gi.repository import SugarGestures
from sugar3.graphics import palettegroup
from sugar3.graphics import animator
from sugar3.graphics import style
from sugar3.graphics.icon import CellRendererIcon


_pointer = None


def _get_pointer_position(widget):
    global _pointer

    if _pointer is None:
        display = widget.get_display()
        manager = display.get_device_manager()
        _pointer = manager.get_client_pointer()
    screen, x, y = _pointer.get_position()
    return (x, y)


def _calculate_gap(a, b):
    """Helper function to find the gap position and size of widget a"""
    # Test for each side if the palette and invoker are
    # adjacent to each other.
    gap = True

    if a.y + a.height == b.y:
        gap_side = Gtk.PositionType.BOTTOM
    elif a.x + a.width == b.x:
        gap_side = Gtk.PositionType.RIGHT
    elif a.x == b.x + b.width:
        gap_side = Gtk.PositionType.LEFT
    elif a.y == b.y + b.height:
        gap_side = Gtk.PositionType.TOP
    else:
        gap = False

    if gap:
        if gap_side == Gtk.PositionType.BOTTOM or \
                gap_side == Gtk.PositionType.TOP:
            gap_start = min(a.width, max(0, b.x - a.x))
            gap_size = max(0, min(a.width,
                                  (b.x + b.width) - a.x) - gap_start)
        elif gap_side == Gtk.PositionType.RIGHT or \
                gap_side == Gtk.PositionType.LEFT:
            gap_start = min(a.height, max(0, b.y - a.y))
            gap_size = max(0, min(a.height,
                                  (b.y + b.height) - a.y) - gap_start)

    if gap and gap_size > 0:
        return (gap_side, gap_start, gap_size)
    else:
        return False


class _PaletteMenuWidget(Gtk.Menu):

    __gtype_name__ = "SugarPaletteMenuWidget"

    __gsignals__ = {
        'enter-notify': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'leave-notify': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self):
        Gtk.Menu.__init__(self)

        accel_group = Gtk.AccelGroup()
        self.sugar_accel_group = accel_group
        self.get_toplevel().add_accel_group(accel_group)

        self._popup_position = (0, 0)
        self._entered = False
        self._mouse_in_palette = False
        self._mouse_in_invoker = False
        self._up = False
        self._invoker = None
        self._menus = []

    def set_accept_focus(self, focus):
        pass

    def get_origin(self):
        res_, x, y = self.get_toplevel().get_window().get_origin()
        return x, y

    def move(self, x, y):
        self._popup_position = (x, y)

    def set_transient_for(self, window):
        pass

    def set_invoker(self, invoker):
        pass

    def _position(self, widget, *data):
        return self._popup_position[0], self._popup_position[1], False

    def popup(self, invoker):
        if self._up:
            return

        # We need to track certain mouse events in order to close the palette
        # when the mouse leaves the palette and the invoker widget, but
        # GtkMenu makes our lives hard here.
        #
        # GtkMenu takes a grab on the root window, meaning that normal
        # enter/leave events are not sent to the relevant widgets.
        # However, connecting enter-notify and leave-notify events in this
        # GtkMenu subclass mean that we get to see the events being grabbed.
        # With certain filtering in place (see _enter_notify_cb and
        # _leave_notify_cb) we are able to accurately determine when the
        # mouse leaves/enters the palette menu. Some spurious events are
        # generated but the important thing is that the last event generated
        # in response to a user action is always reliable (i.e. we will
        # always get a leave event last if the user left the menu,
        # even if we get some strange enter events leading up to it).
        #
        # This is complicated with submenus; in this case the submenu takes
        # the grab, so we must also listen for events on any submenus of
        # the palette and apply the same considerations.
        #
        # The remaining challenge is tracking when the mouse enters or leaves
        # the invoker area. While the appropriate GtkMenu grab is active,
        # we do get informed of such events, however these events will only
        # arrive if the user has entered the menu. If the user hovers over
        # the invoker and then leaves the invoker without entering the palette,
        # we get no enter/leave event.
        # We work around this by tracking mouse motion events. When the mouse
        # moves, we compare the mouse coordinates to the region occupied by the
        # invoker, and this lets us track enter/leave for the invoker widget.

        self._invoker = invoker
        self._find_all_menus(self)
        self.realize()
        for menu in self._menus:
            if self._invoker:
                menu.connect('motion-notify-event', self._motion_notify_cb)
            menu.connect('enter-notify-event', self._enter_notify_cb)
            menu.connect('leave-notify-event', self._leave_notify_cb)
            menu.connect('button-release-event', self._button_release_event_cb)
        self._entered = False
        self._mouse_in_palette = False
        self._mouse_in_invoker = False
        Gtk.Menu.popup(self, None, None, self._position, None, 0, 0)
        self._up = True

    def popdown(self):
        if not self._up:
            return
        Gtk.Menu.popdown(self)

        for menu in self._menus:
            menu.disconnect_by_func(self._motion_notify_cb)
            menu.disconnect_by_func(self._enter_notify_cb)
            menu.disconnect_by_func(self._leave_notify_cb)

        self._up = False
        self._menus = []
        self._invoker = None

    def _find_all_menus(self, menu):
        """
        Recursively find all submenus of menu, adding them to self._menus.
        """
        self._menus.append(menu)
        for child in menu.get_children():
            if not isinstance(child, Gtk.MenuItem):
                continue
            submenu = child.get_submenu()
            if submenu and isinstance(submenu, Gtk.Menu):
                self._find_all_menus(submenu)

    def _enter_notify_cb(self, widget, event):
        if event.mode in (Gdk.CrossingMode.GRAB, Gdk.CrossingMode.GTK_GRAB):
            return False
        if event.get_source_device().get_source() == \
                Gdk.InputSource.TOUCHSCREEN:
            return False
        if Gtk.get_event_widget(event) not in self._menus:
            return False

        self._mouse_in_palette = True
        self._reevaluate_state()
        return False

    def _leave_notify_cb(self, widget, event):
        if event.mode in (Gdk.CrossingMode.GRAB, Gdk.CrossingMode.GTK_GRAB):
            return False
        if event.get_source_device().get_source() == \
                Gdk.InputSource.TOUCHSCREEN:
            return False
        if Gtk.get_event_widget(event) not in self._menus:
            return False

        self._mouse_in_palette = False
        self._reevaluate_state()
        return False

    def _motion_notify_cb(self, widget, event):
        if event.get_source_device().get_source() == \
                Gdk.InputSource.TOUCHSCREEN:
            return False
        x = event.x_root
        y = event.y_root

        rect = self._invoker.get_rect()
        in_invoker = x >= rect.x and x < (rect.x + rect.width) \
            and y >= rect.y and y < (rect.y + rect.height)

        if in_invoker != self._mouse_in_invoker:
            self._mouse_in_invoker = in_invoker
            self._reevaluate_state()

    def _button_release_event_cb(self, widget, event):
        x = event.x_root
        y = event.y_root

        rect = self._invoker.get_rect()
        in_invoker = x >= rect.x and x < (rect.x + rect.width) \
            and y >= rect.y and y < (rect.y + rect.height)

        if in_invoker:
            return True

    def _reevaluate_state(self):
        if self._entered:
            # If we previously advised that the mouse was inside, but now the
            # mouse is outside both the invoker and the palette, notify that
            # the mouse has left.
            if not self._mouse_in_palette and not self._mouse_in_invoker:
                self._entered = False
                self.emit('leave-notify')
        else:
            # If we previously advised that the mouse had left, but now the
            # mouse is inside either the palette or the invoker, notify that
            # the mouse has entered.
            if self._mouse_in_palette or self._mouse_in_invoker:
                self._entered = True
                self.emit('enter-notify')


class _PaletteWindowWidget(Gtk.Window):

    __gtype_name__ = 'SugarPaletteWindowWidget'

    __gsignals__ = {
        'enter-notify': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'leave-notify': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self, palette=None):
        Gtk.Window.__init__(self)

        self._palette = palette
        self.set_decorated(False)
        self.set_resizable(False)
        self.set_position(Gtk.WindowPosition.NONE)

        context = self.get_style_context()
        # Just assume all borders are the same
        border = context.get_border(Gtk.StateFlags.ACTIVE).right
        self.set_border_width(border)

        accel_group = Gtk.AccelGroup()
        self.sugar_accel_group = accel_group
        self.add_accel_group(accel_group)

        self._old_alloc = None
        self._invoker = None
        self._should_accept_focus = True

    def set_accept_focus(self, focus):
        self._should_accept_focus = focus
        if self.get_window() is not None:
            self.get_window().set_accept_focus(focus)

    def get_origin(self):
        res_, x, y = self.get_window().get_origin()
        return x, y

    def do_realize(self):
        Gtk.Window.do_realize(self)

        self.get_window().set_accept_focus(self._should_accept_focus)
        self.set_type_hint(Gdk.WindowTypeHint.DIALOG)

    def do_get_preferred_width(self):
        minimum, natural = Gtk.Window.do_get_preferred_width(self)
        label_width = 0
        if self._palette is not None:
            label_width = self._palette.get_label_width()
        size = max(natural, label_width + 2 * self.get_border_width(),
                   style.GRID_CELL_SIZE * 3)
        return size, size

    def do_size_allocate(self, allocation):
        Gtk.Window.do_size_allocate(self, allocation)

        if self._old_alloc is None or \
           self._old_alloc.x != allocation.x or \
           self._old_alloc.y != allocation.y or \
           self._old_alloc.width != allocation.width or \
           self._old_alloc.height != allocation.height:
            self.queue_draw()

        # We need to store old allocation because when size_allocate
        # is called widget.allocation is already updated.
        # Gtk.Window resizing is different from normal containers:
        # the X window is resized, widget.allocation is updated from
        # the configure request handler and finally size_allocate is called.
        self._old_alloc = allocation

    def set_invoker(self, invoker):
        self._invoker = invoker

    def get_rect(self):
        win_x, win_y = self.get_origin()
        rectangle = self.get_allocation()

        x = win_x + rectangle.x
        y = win_y + rectangle.y
        minimum, natural = self.get_preferred_size()
        rect = Gdk.Rectangle()
        rect.x = x
        rect.y = y
        rect.width = minimum.width
        rect.height = natural.height

        return rect

    def do_draw(self, cr):
        # Fall trough to the container expose handler.
        # (Leaving out the window expose handler which redraws everything)
        Gtk.Window.do_draw(self, cr)

        if self._invoker is not None and self._invoker.has_rectangle_gap():
            invoker = self._invoker.get_rect()
            palette = self.get_rect()
            gap = _calculate_gap(palette, invoker)
        else:
            gap = False

        allocation = self.get_allocation()
        context = self.get_style_context()
        context.add_class('palette')
        if gap:
            cr.save()
            cr.set_source_rgb(0, 0, 0)
            cr.rectangle(0, 0, allocation.width, allocation.height)
            cr.set_line_width(4)
            cr.stroke()
            cr.restore()
            Gtk.render_frame_gap(
                context, cr, 0, 0, allocation.width, allocation.height,
                gap[0], gap[1], gap[1] + gap[2])
        else:
            Gtk.render_frame(
                context, cr, 0, 0, allocation.width, allocation.height)
        return False

    def __enter_notify_event_cb(self, widget, event):
        if event.mode == Gdk.CrossingMode.NORMAL and \
                event.detail != Gdk.NotifyType.INFERIOR:
            self.emit('enter-notify')
        return False

    def __leave_notify_event_cb(self, widget, event):
        if event.mode != Gdk.CrossingMode.NORMAL:
            return False

        if event.detail != Gdk.NotifyType.INFERIOR:
            self.emit('leave-notify')

    def popup(self, invoker):
        if self.get_visible():
            return
        self.connect('enter-notify-event', self.__enter_notify_event_cb)
        self.connect('leave-notify-event', self.__leave_notify_event_cb)
        self.show()

    def popdown(self):
        if not self.get_visible():
            return
        self.disconnect_by_func(self.__enter_notify_event_cb)
        self.disconnect_by_func(self.__leave_notify_event_cb)
        self.hide()


if hasattr(_PaletteWindowWidget, 'set_css_name'):
    _PaletteWindowWidget.set_css_name('palette')


class MouseSpeedDetector(GObject.GObject):

    __gsignals__ = {
        'motion-slow': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'motion-fast': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    _MOTION_SLOW = 1
    _MOTION_FAST = 2

    def __init__(self, delay, thresh):
        """Create MouseSpeedDetector object,
            delay in msec
            threshold in pixels (per tick of 'delay' msec)"""

        GObject.GObject.__init__(self)

        self.parent = None
        self._threshold = thresh
        self._delay = delay
        self._state = None
        self._timeout_hid = None
        self._mouse_pos = None

    def start(self):
        self.stop()

        self._mouse_pos = _get_pointer_position(self.parent)
        self._timeout_hid = GLib.timeout_add(self._delay, self._timer_cb)

    def stop(self):
        if self._timeout_hid is not None:
            GLib.source_remove(self._timeout_hid)
            self._timeout_hid = None
        self._state = None

    def _detect_motion(self):
        oldx, oldy = self._mouse_pos
        (x, y) = _get_pointer_position(self.parent)
        self._mouse_pos = (x, y)

        dist2 = (oldx - x) ** 2 + (oldy - y) ** 2
        if dist2 > self._threshold ** 2:
            return True
        else:
            return False

    def _timer_cb(self):
        motion = self._detect_motion()
        if motion and self._state != self._MOTION_FAST:
            self.emit('motion-fast')
            self._state = self._MOTION_FAST
        elif not motion and self._state != self._MOTION_SLOW:
            self.emit('motion-slow')
            self._state = self._MOTION_SLOW

        return True


class PaletteWindow(GObject.GObject):
    """
    Base class for _ToolbarPalette and Palette.

    Provides basic management of child widget, invoker, and animation.
    """

    __gsignals__ = {
        'popup': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'popdown': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    def __init__(self, **kwargs):
        self._group_id = None
        self._invoker = None
        self._invoker_hids = []
        self._cursor_x = 0
        self._cursor_y = 0
        self._alignment = None
        self._up = False
        self._widget = None

        self._popup_anim = animator.Animator(.5, 10)
        self._popup_anim.add(_PopupAnimation(self))

        self._popdown_anim = animator.Animator(0.6, 10)
        self._popdown_anim.add(_PopdownAnimation(self))

        GObject.GObject.__init__(self, **kwargs)

        self.set_group_id('default')

        self._mouse_detector = MouseSpeedDetector(200, 5)

    def _setup_widget(self):
        self._widget.connect('show', self.__show_cb)
        self._widget.connect('hide', self.__hide_cb)
        self._widget.connect('destroy', self.__destroy_cb)
        self._widget.connect('enter-notify', self.__enter_notify_cb)
        self._widget.connect('leave-notify', self.__leave_notify_cb)
        self._widget.connect('key-press-event', self.__key_press_event_cb)

        self._set_effective_group_id(self._group_id)
        self._widget.set_invoker(self._invoker)

        self._mouse_detector.connect('motion-slow', self._mouse_slow_cb)
        self._mouse_detector.parent = self._widget

    def _teardown_widget(self):
        self._widget.disconnect_by_func(self.__show_cb)
        self._widget.disconnect_by_func(self.__hide_cb)
        self._widget.disconnect_by_func(self.__destroy_cb)
        self._widget.disconnect_by_func(self.__enter_notify_cb)
        self._widget.disconnect_by_func(self.__leave_notify_cb)
        self._widget.disconnect_by_func(self.__key_press_event_cb)
        self._set_effective_group_id(None)

    def destroy(self):
        if self._widget is not None:
            self._widget.destroy()

    def __destroy_cb(self, palette):
        self._mouse_detector.disconnect_by_func(self._mouse_slow_cb)

    def set_invoker(self, invoker):
        for hid in self._invoker_hids[:]:
            self._invoker.disconnect(hid)
            self._invoker_hids.remove(hid)

        self._invoker = invoker
        if self._widget is not None:
            self._widget.set_invoker(invoker)
        if invoker is not None:
            self._invoker_hids.append(self._invoker.connect(
                'mouse-enter', self._invoker_mouse_enter_cb))
            self._invoker_hids.append(self._invoker.connect(
                'mouse-leave', self._invoker_mouse_leave_cb))
            self._invoker_hids.append(self._invoker.connect(
                'right-click', self._invoker_right_click_cb))
            self._invoker_hids.append(self._invoker.connect(
                'toggle-state', self._invoker_toggle_state_cb))

    def get_invoker(self):
        return self._invoker

    invoker = GObject.Property(type=object,
                               getter=get_invoker,
                               setter=set_invoker)

    def _mouse_slow_cb(self, widget):
        self._mouse_detector.stop()
        self._palette_do_popup()

    def _palette_do_popup(self):
        immediate = False

        if self.is_up():
            self._popdown_anim.stop()
            return

        if self._group_id:
            group = palettegroup.get_group(self._group_id)
            if group and group.is_up():
                immediate = True
                group.popdown()

        self.popup(immediate=immediate)

    def is_up(self):
        return self._up

    def _set_effective_group_id(self, group_id):
        if self._group_id:
            group = palettegroup.get_group(self._group_id)
            group.remove(self)
        if group_id:
            group = palettegroup.get_group(group_id)
            group.add(self)

    def set_group_id(self, group_id):
        self._set_effective_group_id(group_id)
        self._group_id = group_id

    def get_group_id(self):
        return self._group_id

    group_id = GObject.Property(type=str,
                                getter=get_group_id,
                                setter=set_group_id)

    def update_position(self):
        invoker = self._invoker
        if invoker is None or self._alignment is None:
            logging.error('Cannot update the palette position.')
            return

        if self._widget is None:
            return

        req = self._widget.size_request()
        if isinstance(self._widget, _PaletteMenuWidget):
            # on Gtk 3.10, menu at the bottom of the screen are resized
            # to not fall out, and report a wrong size.
            # measure the children and move the menu - SL #4673
            total_height = 0
            for child in self._widget.get_children():
                child_req = child.size_request()
                total_height += child_req.height

            # need add the border line width as defined in sugar-artwork
            line_width = 2
            total_height += line_width * 2
            req.height = total_height

        position = invoker.get_position_for_alignment(self._alignment, req)
        if position is None:
            position = invoker.get_position(req)

        self._widget.move(position.x, position.y)

    def get_full_size_request(self):
        return self._widget.size_request()

    def popup(self, immediate=False):
        if self._widget is None:
            return
        if self._invoker is not None:
            full_size_request = self.get_full_size_request()
            self._alignment = self._invoker.get_alignment(full_size_request)

            self.update_position()
            try:
                self._widget.set_transient_for(self._invoker.get_toplevel())
            except TypeError:
                # the expected parent window did likely change e.g. SL #4221
                # popdown the Palette
                self.emit('popdown')
                return

        self._popdown_anim.stop()

        if not immediate:
            self._popup_anim.start()
        else:
            self._popup_anim.stop()
            self._widget.popup(self._invoker)
            # we have to invoke update_position() twice
            # since WM could ignore first move() request
            self.update_position()

    def popdown(self, immediate=False):
        self._popup_anim.stop()
        self._mouse_detector.stop()

        if not immediate:
            self._popdown_anim.start()
        else:
            self._popdown_anim.stop()
            if self._widget is not None:
                self._widget.popdown()

    def on_invoker_enter(self):
        self._popdown_anim.stop()
        self._mouse_detector.start()

    def on_invoker_leave(self):
        self._mouse_detector.stop()
        self.popdown()

    def on_enter(self):
        self._popdown_anim.stop()

    def on_leave(self):
        self.popdown()

    def _invoker_mouse_enter_cb(self, invoker):
        if not self._invoker.locked:
            self.on_invoker_enter()

    def _invoker_mouse_leave_cb(self, invoker):
        if not self._invoker.locked:
            self.on_invoker_leave()

    def _invoker_right_click_cb(self, invoker):
        self.popup(immediate=True)

    def _invoker_toggle_state_cb(self, invoker):
        if self.is_up():
            self.popdown(immediate=True)
        else:
            self.popup(immediate=True)

    def __enter_notify_cb(self, widget):
        if not self._invoker.locked:
            self.on_enter()

    def __leave_notify_cb(self, widget):
        if not self._invoker.locked:
            self.on_leave()

    def __key_press_event_cb(self, window, event):
        if event.keyval == Gdk.KEY_Escape:
            self.popdown()

    def __show_cb(self, widget):
        if self._invoker is not None:
            self._invoker.notify_popup()

        self._up = True
        self.emit('popup')

    def __hide_cb(self, widget):
        if self._invoker:
            self._invoker.notify_popdown()

        self._up = False
        self.emit('popdown')

    def get_rect(self):
        win_x, win_y = self._widget.get_origin()
        rectangle = self._widget.get_allocation()

        x = win_x + rectangle.x
        y = win_y + rectangle.y
        minimum, natural_ = self._widget.get_preferred_size()

        rect = Gdk.Rectangle()
        rect.x = x
        rect.y = y
        rect.width = minimum.width
        rect.height = minimum.height

        return rect


class _PopupAnimation(animator.Animation):

    def __init__(self, palette):
        animator.Animation.__init__(self, 0.0, 1.0)
        self._palette = palette

    def next_frame(self, current):
        if current == 1.0:
            self._palette.popup(immediate=True)


class _PopdownAnimation(animator.Animation):

    def __init__(self, palette):
        animator.Animation.__init__(self, 0.0, 1.0)
        self._palette = palette

    def next_frame(self, current):
        if current == 1.0:
            self._palette.popdown(immediate=True)


class Invoker(GObject.GObject):

    __gtype_name__ = 'SugarPaletteInvoker'

    __gsignals__ = {
        'mouse-enter': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'mouse-leave': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'right-click': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'toggle-state': (GObject.SignalFlags.RUN_FIRST, None, ([])),
        'focus-out': (GObject.SignalFlags.RUN_FIRST, None, ([])),
    }

    ANCHORED = 0
    AT_CURSOR = 1

    BOTTOM = [(0.0, 0.0, 0.0, 1.0), (-1.0, 0.0, 1.0, 1.0)]
    RIGHT = [(0.0, 0.0, 1.0, 0.0), (0.0, -1.0, 1.0, 1.0)]
    TOP = [(0.0, -1.0, 0.0, 0.0), (-1.0, -1.0, 1.0, 0.0)]
    LEFT = [(-1.0, 0.0, 0.0, 0.0), (-1.0, -1.0, 0.0, 1.0)]

    def __init__(self):
        GObject.GObject.__init__(self)

        self.parent = None

        self._screen_area = Gdk.Rectangle()
        self._screen_area.x = self._screen_area.y = 0
        self._screen_area.width = Gdk.Screen.width()
        self._screen_area.height = Gdk.Screen.height()
        self._position_hint = self.ANCHORED
        self._cursor_x = -1
        self._cursor_y = -1
        self._palette = None
        self._cache_palette = True
        self._toggle_palette = False
        self._lock_palette = False
        self.locked = False

    def attach(self, parent):
        self.parent = parent

    def detach(self):
        self.parent = None
        if self._palette is not None:
            self._palette.destroy()
            self._palette = None

    def _get_position_for_alignment(self, alignment, palette_dim):
        palette_halign = alignment[0]
        palette_valign = alignment[1]
        invoker_halign = alignment[2]
        invoker_valign = alignment[3]

        if self._cursor_x == -1 or self._cursor_y == -1:
            position = _get_pointer_position(self.parent)
            (self._cursor_x, self._cursor_y) = position

        if self._position_hint is self.ANCHORED:
            rect = self.get_rect()
        else:
            dist = style.PALETTE_CURSOR_DISTANCE
            rect = Gdk.Rectangle()
            rect.x = self._cursor_x - dist
            rect.y = self._cursor_y - dist
            rect.width = rect.height = dist * 2

        palette_width, palette_height = palette_dim.width, palette_dim.height

        x = rect.x + rect.width * invoker_halign + \
            palette_width * palette_halign

        y = rect.y + rect.height * invoker_valign + \
            palette_height * palette_valign

        rect = Gdk.Rectangle()
        rect.x = int(x)
        rect.y = int(y)
        rect.width = palette_width
        rect.height = palette_height
        return rect

    def _in_screen(self, rect):
        return rect.x >= self._screen_area.x and \
            rect.y >= self._screen_area.y and \
            rect.x + rect.width <= self._screen_area.width and \
            rect.y + rect.height <= self._screen_area.height

    def _get_area_in_screen(self, rect):
        """Return area of rectangle visible in the screen"""

        x1 = max(rect.x, self._screen_area.x)
        y1 = max(rect.y, self._screen_area.y)
        x2 = min(rect.x + rect.width,
                 self._screen_area.x + self._screen_area.width)
        y2 = min(rect.y + rect.height,
                 self._screen_area.y + self._screen_area.height)

        return (x2 - x1) * (y2 - y1)

    def _get_alignments(self):
        if self._position_hint is self.AT_CURSOR:
            return [(0.0, 0.0, 1.0, 1.0),
                    (0.0, -1.0, 1.0, 0.0),
                    (-1.0, -1.0, 0.0, 0.0),
                    (-1.0, 0.0, 0.0, 1.0)]
        else:
            return self.BOTTOM + self.RIGHT + self.TOP + self.LEFT

    def get_position_for_alignment(self, alignment, palette_dim):
        rect = self._get_position_for_alignment(alignment, palette_dim)
        if self._in_screen(rect):
            return rect
        else:
            return None

    def get_position(self, palette_dim):
        alignment = self.get_alignment(palette_dim)
        rect = self._get_position_for_alignment(alignment, palette_dim)

        # In case our efforts to find an optimum place inside the screen
        # failed, just make sure the palette fits inside the screen if at all
        # possible.
        rect.x = max(0, rect.x)
        rect.y = max(0, rect.y)

        rect.x = min(rect.x, self._screen_area.width - rect.width)
        rect.y = min(rect.y, self._screen_area.height - rect.height)

        return rect

    def get_alignment(self, palette_dim):
        best_alignment = None
        best_area = -1
        for alignment in self._get_alignments():
            pos = self._get_position_for_alignment(alignment, palette_dim)
            if self._in_screen(pos):
                return alignment

            area = self._get_area_in_screen(pos)
            if area > best_area:
                best_alignment = alignment
                best_area = area

        # Palette horiz/vert alignment
        ph = best_alignment[0]
        pv = best_alignment[1]

        # Invoker horiz/vert alignment
        ih = best_alignment[2]
        iv = best_alignment[3]

        rect = self.get_rect()
        screen_area = self._screen_area

        if best_alignment in self.LEFT or best_alignment in self.RIGHT:
            dtop = rect.y - screen_area.y
            dbottom = screen_area.y + screen_area.height - rect.y - rect.width

            iv = 0

            # Set palette_valign to align to screen on the top
            if dtop > dbottom:
                pv = -float(dtop) / palette_dim.height

            # Set palette_valign to align to screen on the bottom
            else:
                pv = -float(palette_dim.height - dbottom - rect.height) \
                    / palette_dim.height

        elif best_alignment in self.TOP or best_alignment in self.BOTTOM:
            dleft = rect.x - screen_area.x
            dright = screen_area.x + screen_area.width - rect.x - rect.width

            ih = 0

            if palette_dim.width == 0:
                ph = 0

            else:
                # Set palette_halign to align to screen on left
                if dleft > dright:
                    ph = -float(dleft) / palette_dim.width

                # Set palette_halign to align to screen on right
                else:
                    ph = -float(palette_dim.width - dright - rect.width) \
                        / palette_dim.width

        return (ph, pv, ih, iv)

    def has_rectangle_gap(self):
        return False

    def draw_rectangle(self, event, palette):
        pass

    def notify_popup(self):
        pass

    def notify_popdown(self):
        self._cursor_x = -1
        self._cursor_y = -1

    def _ensure_palette_exists(self):
        if self.parent and self.palette is None:
            palette = self.parent.create_palette()
            if palette is not None:
                self.palette = palette

    def notify_mouse_enter(self):
        self._ensure_palette_exists()
        self.emit('mouse-enter')

    def notify_mouse_leave(self):
        self.emit('mouse-leave')

    def notify_right_click(self, x=None, y=None):
        '''
        Notify the palette invoker of a right click and expand the
        palette as required.  The x and y args should be that of
        where the event happened, relative to the root of the screen.

        Args
            x (float): the x coord of the event relative to the root
                of the screen, eg. :class:`Gdk.EventTouch.x_root`
            y (float): the y coord of the event relative to the root
                of the screen, eg. :class:`Gdk.EventTouch.y_root`
        '''
        self._ensure_palette_exists()
        self._process_event(x, y)
        self.emit('right-click')

    def notify_toggle_state(self):
        self._ensure_palette_exists()
        self.emit('toggle-state')

    def _process_event(self, x, y):
        if x is None or y is None:
            return

        self._cursor_x = x
        self._cursor_y = y

    def get_palette(self):
        return self._palette

    def set_palette(self, palette):
        if self._palette is not None:
            self._palette.popdown(immediate=True)
            self._palette.props.invoker = None
            # GTK pops down the palette before it invokes the actions on the
            # menu item. We need to postpone destruction of the palette until
            # after all signals have propagated from the menu item to the
            # palette owner.
            GLib.idle_add(lambda old_palette=self._palette:
                          old_palette.destroy(),
                          priority=GLib.PRIORITY_LOW)

        self._palette = palette

        if self._palette is not None:
            self._palette.props.invoker = self
            self._palette.connect('popdown', self.__palette_popdown_cb)

    palette = GObject.Property(
        type=object, setter=set_palette, getter=get_palette)

    def get_cache_palette(self):
        return self._cache_palette

    def set_cache_palette(self, cache_palette):
        self._cache_palette = cache_palette

    cache_palette = GObject.Property(type=object, setter=set_cache_palette,
                                     getter=get_cache_palette)
    """Whether the invoker will cache the palette after its creation. Defaults
    to True.
    """

    def get_toggle_palette(self):
        return self._toggle_palette

    def set_toggle_palette(self, toggle_palette):
        self._toggle_palette = toggle_palette

    toggle_palette = GObject.Property(type=object, setter=set_toggle_palette,
                                      getter=get_toggle_palette)
    """Whether the invoker will popup/popdown the Palette on
    button left click/touch tap. Defaults to False.
    """

    def get_lock_palette(self):
        return self._lock_palette

    def set_lock_palette(self, lock_palette):
        self._lock_palette = lock_palette

    lock_palette = GObject.Property(type=object, setter=set_lock_palette,
                                    getter=get_lock_palette)
    """Whether the invoker will lock the Palette and
    ignore mouse events. Defaults to False.
    """

    def __palette_popdown_cb(self, palette):
        if not self.props.cache_palette:
            self.set_palette(None)

    def primary_text_clicked(self):
        """Implemented by invokers that can be clicked"""
        pass


class WidgetInvoker(Invoker):

    def __init__(self, parent=None, widget=None):
        Invoker.__init__(self)

        self._widget = None
        self._expanded = False
        self._pointer_position = (-1, -1)
        self._enter_hid = None
        self._leave_hid = None
        self._release_hid = None
        self._click_hid = None
        self._touch_hid = None
        self._draw_hid = None
        self._long_pressed_recognized = False
        self._long_pressed_hid = None
        self._long_pressed_controller = SugarGestures.LongPressController()

        if parent or widget:
            self.attach_widget(parent, widget)

    def attach_widget(self, parent, widget=None):
        if widget:
            self._widget = widget
        else:
            self._widget = parent

        self._pointer_position = _get_pointer_position(self._widget)

        self.notify('widget')

        self._enter_hid = self._widget.connect('enter-notify-event',
                                               self.__enter_notify_event_cb)
        self._leave_hid = self._widget.connect('leave-notify-event',
                                               self.__leave_notify_event_cb)
        if GObject.signal_lookup('clicked', self._widget) != 0:
            self._click_hid = self._widget.connect('clicked',
                                                   self.__click_event_cb)
        self._touch_hid = self._widget.connect('touch-event',
                                               self.__touch_event_cb)
        self._release_hid = \
            self._widget.connect('button-release-event',
                                 self.__button_release_event_cb)
        self._draw_hid = self._widget.connect_after('draw', self.__drawing_cb)

        self._long_pressed_hid = self._long_pressed_controller.connect(
            'pressed', self.__long_pressed_event_cb, self._widget)
        self._long_pressed_controller.attach(
            self._widget,
            SugarGestures.EventControllerFlags.NONE)

        self.attach(parent)

    def detach(self):
        Invoker.detach(self)
        self._widget.disconnect(self._enter_hid)
        self._widget.disconnect(self._leave_hid)
        self._widget.disconnect(self._release_hid)
        self._widget.disconnect(self._draw_hid)
        if self._click_hid:
            self._widget.disconnect(self._click_hid)
        self._widget.disconnect(self._touch_hid)
        self._long_pressed_controller.detach(self._widget)
        self._long_pressed_controller.disconnect(self._long_pressed_hid)

    def get_rect(self):
        allocation = self._widget.get_allocation()
        window = self._widget.get_window()
        if window is not None:
            res_, x, y = window.get_origin()
        else:
            logging.warning(
                "Trying to position palette with invoker that's not realized.")
            x = 0
            y = 0

        if not self._widget.get_has_window():
            x += allocation.x
            y += allocation.y

        rect = Gdk.Rectangle()
        rect.x = x
        rect.y = y
        rect.width = allocation.width
        rect.height = allocation.height
        return rect

    def has_rectangle_gap(self):
        return True

    def draw_rectangle(self, cr, palette):
        allocation = self.parent.get_allocation()

        context = self.parent.get_style_context()
        context.add_class('toolitem')
        context.add_class('palette-down')

        gap = _calculate_gap(self.get_rect(), palette.get_rect())
        if gap:
            Gtk.render_frame_gap(context, cr, 0, 0,
                                 allocation.width,
                                 allocation.height,
                                 gap[0], gap[1], gap[1] + gap[2])

    def __enter_notify_event_cb(self, widget, event):
        if (event.x_root, event.y_root) == self._pointer_position:
            self._pointer_position = (-1, -1)
            return False
        if event.mode == Gdk.CrossingMode.NORMAL:
            self.notify_mouse_enter()

    def __leave_notify_event_cb(self, widget, event):
        if event.mode == Gdk.CrossingMode.NORMAL:
            self.notify_mouse_leave()

    def __touch_event_cb(self, button, event):
        if event.type == Gdk.EventType.TOUCH_END:
            if self._long_pressed_recognized:
                self._long_pressed_recognized = False
                return True
        return False

    def __click_event_cb(self, button):
        event = Gtk.get_current_event()
        if not event:
            # not an event from a user interaction, this can be when
            # the clicked event is emitted on a 'active' property
            # change of ToggleToolButton for example
            return
        if event and button != Gtk.get_event_widget(event):
            # another special case for the ToggleToolButton: this handles
            # the case where we select an item and the active property
            # of the other one changes to 'False'
            return

        if self.props.lock_palette and not self.locked:
            self.locked = True
            if hasattr(self.parent, 'set_expanded'):
                self.parent.set_expanded(True)

        if self.props.toggle_palette:
            self.notify_toggle_state()

    def __button_release_event_cb(self, widget, event):
        if event.button == 1 and not self._click_hid:
            if self.props.lock_palette and not self.locked:
                self.locked = True
                if hasattr(self.parent, 'set_expanded'):
                    self.parent.set_expanded(True)

            if self.props.toggle_palette:
                self.notify_toggle_state()
        elif event.button == 3:
            self.notify_right_click(event.x_root, event.y_root)
            return True
        else:
            return False

    def __long_pressed_event_cb(self, controller, x, y, widget):
        self._long_pressed_recognized = True
        self.notify_right_click(x, y)

    def get_toplevel(self):
        return self._widget.get_toplevel()

    def notify_popup(self):
        Invoker.notify_popup(self)
        self._widget.queue_draw()

    def notify_popdown(self):
        self.locked = False
        Invoker.notify_popdown(self)
        self._widget.queue_draw()

    def _get_widget(self):
        return self._widget
    widget = GObject.Property(type=object, getter=_get_widget, setter=None)

    def __drawing_cb(self, widget, cr):
        if not self.props.lock_palette:
            return False
        alloc = widget.get_allocation()
        arrow_size = style.TOOLBAR_ARROW_SIZE / 2
        y = alloc.height - arrow_size
        x = (alloc.width - arrow_size) / 2
        context = widget.get_style_context()
        context.add_class('toolitem')
        if self.locked:
            Gtk.render_arrow(context, cr, 0, x, y, arrow_size)
        else:
            Gtk.render_arrow(context, cr, math.pi, x, y, arrow_size)


class CursorInvoker(Invoker):

    def __init__(self, parent=None):
        Invoker.__init__(self)

        self._position_hint = self.AT_CURSOR
        self._pointer_position = (-1, -1)
        self._enter_hid = None
        self._leave_hid = None
        self._release_hid = None
        self._item = None
        self._long_pressed_recognized = False
        self._long_pressed_hid = None
        self._long_pressed_controller = SugarGestures.LongPressController()

        if parent:
            self.attach(parent)

    def attach(self, parent):
        Invoker.attach(self, parent)

        self._item = parent

        self._pointer_position = _get_pointer_position(self.parent)

        self._enter_hid = self._item.connect('enter-notify-event',
                                             self.__enter_notify_event_cb)
        self._leave_hid = self._item.connect('leave-notify-event',
                                             self.__leave_notify_event_cb)
        self._release_hid = self._item.connect('button-release-event',
                                               self.__button_release_event_cb)
        self._long_pressed_hid = self._long_pressed_controller.connect(
            'pressed',
            self.__long_pressed_event_cb, self._item)
        self._long_pressed_controller.attach(
            self._item,
            SugarGestures.EventControllerFlags.NONE)

    def detach(self):
        Invoker.detach(self)
        self._item.disconnect_by_func(self.__enter_notify_event_cb)
        self._item.disconnect_by_func(self.__leave_notify_event_cb)
        self._item.disconnect_by_func(self.__button_release_event_cb)
        self._long_pressed_controller.detach(self._item)
        self._long_pressed_controller.disconnect(self._long_pressed_hid)

    def get_default_position(self):
        return self.AT_CURSOR

    def get_rect(self):
        window = self._item.get_window()
        allocation = self._item.get_allocation()
        rect = Gdk.Rectangle()
        rect.x, rect.y = window.get_root_coords(allocation.x, allocation.y)
        rect.width = allocation.width
        rect.height = allocation.height
        return rect

    def __enter_notify_event_cb(self, button, event):
        if (event.x_root, event.y_root) == self._pointer_position:
            self._pointer_position = (-1, -1)
            return False
        if event.mode == Gdk.CrossingMode.NORMAL:
            self.notify_mouse_enter()
        return False

    def __leave_notify_event_cb(self, button, event):
        if event.mode == Gdk.CrossingMode.NORMAL:
            self.notify_mouse_leave()
        return False

    def __button_release_event_cb(self, button, event):
        # check if the release is done outside of the parent widget
        alloc = self._item.get_allocation()
        if not (0 < event.x < alloc.width and 0 < event.y < alloc.height):
            return False

        if self._long_pressed_recognized:
            self._long_pressed_recognized = False
            return True
        if event.button == 1:
            if self.props.toggle_palette:
                self.notify_toggle_state()
        if event.button == 3:
            self.notify_right_click(event.x_root, event.y_root)
            return True
        else:
            return False

    def __long_pressed_event_cb(self, controller, x, y, widget):
        self._long_pressed_recognized = True
        x, y = widget.get_window().get_root_coords(x, y)
        self.notify_right_click(x, y)

    def get_toplevel(self):
        return self._item.get_toplevel()


class ToolInvoker(WidgetInvoker):
    '''
    A palette invoker for toolbar buttons and other items.  This invoker
    will properly align the palette so that is perpendicular to the toolbar
    (a horizontal toolbar will spawn a palette going downwards).  It also
    draws the highlights specific to a toolitem.

    For :class:`sugar3.graphics.toolbutton.ToolButton` and subclasses, you
    should not use the toolinvoker directly.  Instead, just subclass the
    tool button and override the `create_palette` function.

    Args:
        parent (Gtk.Widget):  toolitem to connect invoker to
    '''

    def __init__(self, parent=None):
        WidgetInvoker.__init__(self)
        self._tool = None

        if parent:
            self.attach_tool(parent)

    def attach_tool(self, widget):
        '''
        Attach a toolitem to the invoker.  Same behaviour as passing the
        `parent` argument to the constructor.

        Args:
            widget (Gtk.Widget):  toolitem to connect invoker to
        '''
        self._tool = widget
        self.attach_widget(widget, widget.get_child())

    def _get_alignments(self):
        parent = self._widget.get_parent()
        if parent is None:
            return WidgetInvoker._get_alignments(self)

        if parent.get_orientation() is Gtk.Orientation.HORIZONTAL:
            return self.BOTTOM + self.TOP
        else:
            return self.LEFT + self.RIGHT

    def primary_text_clicked(self):
        self._widget.emit('clicked')

    def notify_popup(self):
        WidgetInvoker.notify_popup(self)
        self._tool.queue_draw()

    def notify_popdown(self):
        WidgetInvoker.notify_popdown(self)
        self._tool.queue_draw()


class TreeViewInvoker(Invoker):
    def __init__(self):
        Invoker.__init__(self)

        self._tree_view = None
        self._motion_hid = None
        self._release_hid = None
        self._long_pressed_hid = None
        self._position_hint = self.AT_CURSOR

        self._long_pressed_controller = SugarGestures.LongPressController()

        self._tree_view = None
        self._path = None
        self._column = None

        self.palette = None

    def attach_treeview(self, tree_view):
        self._tree_view = tree_view

        self._motion_hid = tree_view.connect('motion-notify-event',
                                             self.__motion_notify_event_cb)
        self._release_hid = tree_view.connect('button-release-event',
                                              self.__button_release_event_cb)
        self._long_pressed_hid = self._long_pressed_controller.connect(
            'pressed', self.__long_pressed_event_cb, tree_view)
        self._long_pressed_controller.attach(
            tree_view,
            SugarGestures.EventControllerFlags.NONE)

        Invoker.attach(self, tree_view)

    def detach(self):
        Invoker.detach(self)
        self._tree_view.disconnect(self._motion_hid)
        self._tree_view.disconnect(self._release_hid)
        self._long_pressed_controller.detach(self._tree_view)
        self._long_pressed_controller.disconnect(self._long_pressed_hid)

    def get_rect(self):
        return self._tree_view.get_background_area(self._path, self._column)

    def get_toplevel(self):
        return self._tree_view.get_toplevel()

    def __motion_notify_event_cb(self, widget, event):
        here = self._tree_view.get_path_at_pos(int(event.x), int(event.y))
        if here is None:
            if self._path is not None:
                self.notify_mouse_leave()
            self._path = None
            self._column = None
            return

        path, column, x_, y_ = here
        if path != self._path or column != self._column:
            self._redraw_cell(self._path, self._column)
            self._redraw_cell(path, column)

            self._path = path
            self._column = column

            if self.palette is not None:
                self.palette.popdown(immediate=True)
                self.palette = None

            self.notify_mouse_enter()

    def _redraw_cell(self, path, column):
        area = self._tree_view.get_background_area(path, column)
        x, y = \
            self._tree_view.convert_bin_window_to_widget_coords(area.x, area.y)
        self._tree_view.queue_draw_area(x, y, area.width, area.height)

    def __button_release_event_cb(self, widget, event):
        x, y = int(event.x), int(event.y)
        here = self._tree_view.get_path_at_pos(x, y)
        if here is None:
            return False
        path, column, cell_x, cell_y = here
        self._path = path
        self._column = column
        if event.button == 1:
            # left mouse button
            if self.palette is not None:
                self.palette.popdown(immediate=True)
            # NOTE: we don't use columns with more than one cell
            cellrenderer = column.get_cells()[0]
            if cellrenderer is not None and \
                    isinstance(cellrenderer, CellRendererIcon):
                cellrenderer.emit('clicked', path)
            # So the treeview receives it and knows a drag isn't going on
            return False
        elif event.button == 3:
            # right mouse button
            self._ensure_palette_exists()
            self.notify_right_click(event.x_root, event.y_root)
            return True
        else:
            return False

    def __long_pressed_event_cb(self, controller, x, y, widget):
        path, column, x_, y_ = self._tree_view.get_path_at_pos(x, y)
        self._path = path
        self._column = column
        self._ensure_palette_exists()
        self.notify_right_click(x, y)

    def _ensure_palette_exists(self):
        if hasattr(self._tree_view, 'create_palette'):
            self.palette = self._tree_view.create_palette(
                self._path, self._column)
        else:
            self.palette = None

    def notify_popdown(self):
        Invoker.notify_popdown(self)
        self.palette = None



# Copyright (C) 2006-2008, Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

import os
import sys
import six
import logging

# Change the default encoding to avoid UnicodeDecodeError
# http://lists.sugarlabs.org/archive/sugar-devel/2012-August/038928.html
if six.PY2:
    reload(sys)
    sys.setdefaultencoding('utf-8')

import gettext
from argparse import ArgumentParser

import dbus
import dbus.service
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from sugar3.activity import activityhandle
from sugar3 import config
from sugar3.bundle.activitybundle import ActivityBundle
from sugar3 import logger

from sugar3.bundle.bundle import MalformedBundleException

from errno import EEXIST

import time
import hashlib
import random


def _makedirs(path):
    try:
        os.makedirs(path)
    except OSError as e:
        if e.errno != EEXIST:
            raise e


def create_activity_instance(constructor, handle):
    activity = constructor(handle)
    activity.show()
    return activity


def get_single_process_name(bundle_id):
    return bundle_id


def get_single_process_path(bundle_id):
    return '/' + bundle_id.replace('.', '/')


class SingleProcess(dbus.service.Object):

    def __init__(self, name_service, constructor):
        self.constructor = constructor

        bus = dbus.SessionBus()
        bus_name = dbus.service.BusName(name_service, bus=bus)
        object_path = get_single_process_path(name_service)
        dbus.service.Object.__init__(self, bus_name, object_path)

    @dbus.service.method('org.laptop.SingleProcess', in_signature='a{sv}')
    def create(self, handle_dict):
        handle = activityhandle.create_from_dict(handle_dict)
        create_activity_instance(self.constructor, handle)


def main():
    usage = 'usage: %prog [options] [activity dir] [python class]'
    epilog = 'If you are running from a directory containing an Activity, ' \
             'the argument may be omitted.  Otherwise please provide either '\
             'a directory containing a Sugar Activity [activity dir], a '\
             '[python_class], or both.'

    parser = ArgumentParser(usage=usage, epilog=epilog)
    parser.add_argument('-b', '--bundle-id', dest='bundle_id',
                        help='identifier of the activity bundle')
    parser.add_argument('-a', '--activity-id', dest='activity_id',
                        help='identifier of the activity instance')
    parser.add_argument('-o', '--object-id', dest='object_id',
                        help='identifier of the associated datastore object')
    parser.add_argument('-u', '--uri', dest='uri',
                        help='URI to load')
    parser.add_argument('-s', '--single-process', dest='single_process',
                        action='store_true',
                        help='start all the instances in the same process')
    parser.add_argument('-i', '--invited', dest='invited',
                        action='store_true', default=False,
                        help='the activity is being launched for handling an '
                             'invite from the network')
    
    options, args = parser.parse_known_args()

    logger.start()

    activity_class = None
    if len(args) == 2:
        activity_class = args[1]
        os.chdir(args[0])
    elif len(args) == 1:
        if os.path.isdir(args[0]):
            os.chdir(args[0])
        else:
            activity_class = args[0]

    bundle_path = os.path.abspath(os.curdir)
    sys.path.insert(0, bundle_path)

    try:
        bundle = ActivityBundle(bundle_path)
    except MalformedBundleException:
        parser.print_help()
        exit(0)

    if not activity_class:
        command = bundle.get_command()
        if command.startswith('sugar-activity'):
            if not command.startswith('sugar-activity3'):
                logging.warning("Activity written for Python 2,"
                                " consider porting to Python 3.")
            activity_class = command.split(" ")[1]

    # when an activity is started outside sugar,
    # activityfactory.get_environment has not executed in parent
    # process, so parts of get_environment must happen here.
    if 'SUGAR_BUNDLE_PATH' not in os.environ:
        profile_id = os.environ.get('SUGAR_PROFILE', 'default')
        home_dir = os.environ.get('SUGAR_HOME', os.path.expanduser('~/.sugar'))
        base = os.path.join(home_dir, profile_id)
        activity_root = os.path.join(base, bundle.get_bundle_id())

        instance_dir = os.path.join(activity_root, 'instance')
        _makedirs(instance_dir)

        data_dir = os.path.join(activity_root, 'data')
        _makedirs(data_dir)

        tmp_dir = os.path.join(activity_root, 'tmp')
        _makedirs(tmp_dir)

        os.environ['SUGAR_BUNDLE_PATH'] = bundle_path
        os.environ['SUGAR_BUNDLE_ID'] = bundle.get_bundle_id()
        os.environ['SUGAR_ACTIVITY_ROOT'] = activity_root

    os.environ['SUGAR_BUNDLE_NAME'] = bundle.get_name()
    os.environ['SUGAR_BUNDLE_VERSION'] = str(bundle.get_activity_version())

    # must be done early, some activities set translations globally, SL #3654
    activity_locale_path = os.environ.get("SUGAR_LOCALEDIR",
                                          config.locale_path)

    gettext.bindtextdomain(bundle.get_bundle_id(), activity_locale_path)
    gettext.bindtextdomain('sugar-toolkit-gtk3', config.locale_path)
    gettext.textdomain(bundle.get_bundle_id())

    splitted_module = activity_class.rsplit('.', 1)
    module_name = splitted_module[0]
    class_name = splitted_module[1]

    module = __import__(module_name)
    for comp in module_name.split('.')[1:]:
        module = getattr(module, comp)

    activity_constructor = getattr(module, class_name)

    if not options.activity_id:
        # Generate random hash
        data = '%s%s' % (time.time(), random.randint(10000, 100000))
        random_hash = hashlib.sha1(data.encode()).hexdigest()
        options.activity_id = random_hash
        options.bundle_id = bundle.get_bundle_id()

    activity_handle = activityhandle.ActivityHandle(
        activity_id=options.activity_id,
        object_id=options.object_id, uri=options.uri,
        invited=options.invited)

    if options.single_process is True:
        sessionbus = dbus.SessionBus()

        service_name = get_single_process_name(options.bundle_id)
        service_path = get_single_process_path(options.bundle_id)

        bus_object = sessionbus.get_object(
            'org.freedesktop.DBus', '/org/freedesktop/DBus')
        try:
            name = bus_object.GetNameOwner(
                service_name, dbus_interface='org.freedesktop.DBus')
        except dbus.DBusException:
            name = None

        if not name:
            SingleProcess(service_name, activity_constructor)
        else:
            try:
                single_process = sessionbus.get_object(service_name,
                                                       service_path)
                single_process.create(
                    activity_handle.get_dict(),
                    dbus_interface='org.laptop.SingleProcess')

                print('Created %s in a single process.' % service_name)
                sys.exit(0)
            except (TypeError, dbus.DBusException):
                print('Could not communicate with the instance process,'
                      'launching a new process')

    if hasattr(module, 'start'):
        module.start()

    instance = create_activity_instance(activity_constructor, activity_handle)

    if hasattr(instance, 'run_main_loop'):
        instance.run_main_loop()



# Copyright (C) 2007, Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

'''
The style module defines constants for spacing and sizing, as well as
classes for  Colors and Fonts. Text rendering is handled by Pango and
colors are inputted by their HTML code (e.g. #FFFFFFFF)

All the constants are expressed in pixels. They are defined for the XO
screen and are usually adapted to different resolution by applying a
zoom factor.
'''

import os
import logging

from gi.repository import Gdk
from gi.repository import Pango
from gi.repository import Gio


FOCUS_LINE_WIDTH = 2
_TAB_CURVATURE = 1
ELLIPSIZE_MODE_DEFAULT = Pango.EllipsizeMode.END


def _compute_zoom_factor():
    '''
    Calculates zoom factor based on size of screen.
    Returns double representing fraction of maximum possible screen size
    '''
    try:
        scaling = int(os.environ.get('SUGAR_SCALING', '100'))
        return scaling / 100.0
    except ValueError:
        logging.error('Invalid SUGAR_SCALING.')

    return 1.0


class Font(object):
    '''
    A font defines the style of how the text should be rendered.

    Args:
        desc (str): a description of the Font object
    '''

    def __init__(self, desc):
        self._desc = desc

    def __str__(self):
        '''
        Returns description of font
        '''
        return self._desc

    def get_pango_desc(self):
        '''
        Returns Pango description of font
        '''
        return Pango.FontDescription(self._desc)


class Color(object):
    '''
    A Color object defines a specific color.

    Args:
        color (str): String in the form #FFFFFF representing the color

        alpha (double):  transparency of color
    '''

    def __init__(self, color, alpha=1.0):
        self._r, self._g, self._b = self._html_to_rgb(color)
        self._a = alpha

    def get_rgba(self):
        '''
        Returns 4-tuple of red, green, blue, and alpha levels in range 0-1
        '''
        return (self._r, self._g, self._b, self._a)

    def get_int(self):
        '''
        Returns color encoded as an int, in the form rgba
        '''
        return int(self._a * 255) + (int(self._b * 255) << 8) + \
            (int(self._g * 255) << 16) + (int(self._r * 255) << 24)

    def get_gdk_color(self):
        '''
        Returns GDK standard color
        '''
        return Gdk.Color(int(self._r * 65535), int(self._g * 65535),
                         int(self._b * 65535))

    def get_html(self):
        '''
        Returns string in the standard html Color format (#FFFFFF)
        '''
        return '#%02x%02x%02x' % (
            int(self._r * 255), int(self._g * 255), int(self._b * 255))

    def _html_to_rgb(self, html_color):
        '''
        Converts and returns (r, g, b) tuple in float format from
        standard HTML format (#FFFFFF).  Colors will be in range 0-1

        Args:
            html_color (string): html string in the format #FFFFFF
        '''

        html_color = html_color.strip()
        if html_color[0] == '#':
            html_color = html_color[1:]
        if len(html_color) != 6:
            raise ValueError('input #%s is not in #RRGGBB format' % html_color)

        r, g, b = html_color[:2], html_color[2:4], html_color[4:]
        r, g, b = [int(n, 16) for n in (r, g, b)]
        r, g, b = (r / 255.0, g / 255.0, b / 255.0)

        return (r, g, b)

    def get_svg(self):
        '''
        Returns HTML formatted color, unless the color is completely
        transparent, in which case returns "none"
        '''

        if self._a == 0.0:
            return 'none'
        else:
            return self.get_html()


def zoom(units):
    '''
    Returns size of units pixels at current zoom level

    Args:
        units (int or float): size of item at full size
    '''
    return int(ZOOM_FACTOR * units)


ZOOM_FACTOR = _compute_zoom_factor()  #: scale factor, as float (eg. 0.72, 1.0)

DEFAULT_SPACING = zoom(15)  #: Spacing is placed in-between elements
DEFAULT_PADDING = zoom(6)  #: Padding is placed around an element

#: allow elements to tile neatly within boundaries of a grid
#: http://wiki.sugarlabs.org/go/Human_Interface_Guidelines#The_Grid_System
GRID_CELL_SIZE = zoom(75)

LINE_WIDTH = zoom(2)  #: Thickness of a separator line

#: icon that fits within a grid cell
STANDARD_ICON_SIZE = zoom(55)
#: small icon, used in palette menu items
SMALL_ICON_SIZE = zoom(33)
#: larger than standard
MEDIUM_ICON_SIZE = zoom(55 * 1.5)
#: larger than medium, used in journal empty view
LARGE_ICON_SIZE = zoom(55 * 2.0)
#: larger than large, used in activity pulsing launcher icon
XLARGE_ICON_SIZE = zoom(55 * 2.75)

if 'org.sugarlabs.font' in Gio.Settings.list_schemas():
    settings = Gio.Settings('org.sugarlabs.font')
    #: User's preferred font size
    FONT_SIZE = settings.get_double('default-size')
    #: User's preferred font face
    FONT_FACE = settings.get_string('default-face')
else:
    #: User's preferred font size
    FONT_SIZE = 10
    #: User's preferred font face
    FONT_FACE = 'Sans Serif'

#: Normal font
FONT_NORMAL = Font('%s %f' % (FONT_FACE, FONT_SIZE))
#: Bold font
FONT_BOLD = Font('%s bold %f' % (FONT_FACE, FONT_SIZE))
#: Height in pixels of normal font
FONT_NORMAL_H = zoom(24)
#: Height in pixels of bold font
FONT_BOLD_H = zoom(24)

# old style toolbox design
TOOLBOX_SEPARATOR_HEIGHT = zoom(9)
TOOLBOX_HORIZONTAL_PADDING = zoom(75)
TOOLBOX_TAB_VBORDER = int((zoom(36) - FONT_NORMAL_H - FOCUS_LINE_WIDTH) / 2)
TOOLBOX_TAB_HBORDER = zoom(15) - FOCUS_LINE_WIDTH - _TAB_CURVATURE
TOOLBOX_TAB_LABEL_WIDTH = zoom(150 - 15 * 2)

COLOR_BLACK = Color('#000000')  #: Black
COLOR_WHITE = Color('#FFFFFF')  #: White
#: Fully transparent color
COLOR_TRANSPARENT = Color('#FFFFFF', alpha=0.0)
#: Default background color of a window
COLOR_PANEL_GREY = Color('#C0C0C0')
#: Background color of selected entry
COLOR_SELECTION_GREY = Color('#A6A6A6')
#: Color of toolbars
COLOR_TOOLBAR_GREY = Color('#282828')
#: Color of buttons
COLOR_BUTTON_GREY = Color('#808080')
#: Fill colour of an inactive button
COLOR_INACTIVE_FILL = Color('#9D9FA1')
#: Stroke colour of an inactive button
COLOR_INACTIVE_STROKE = Color('#757575')
#: Background color of entry
COLOR_TEXT_FIELD_GREY = Color('#E5E5E5')
#: Color of highlighted text
COLOR_HIGHLIGHT = Color('#E7E7E7')

#: Cursor invoked palettes will be placed this far from the cursor
PALETTE_CURSOR_DISTANCE = zoom(10)

#: Size of arrow displayed under toolbar buttons to represent their status
TOOLBAR_ARROW_SIZE = zoom(24)

#: Max width of text in a palette menu, in chars
MENU_WIDTH_CHARS = 60



# Copyright (C) 2007, One Laptop Per Child
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
STABLE.
"""

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import Gdk

from sugar3.graphics import style
from sugar3.graphics.palette import ToolInvoker
from sugar3.graphics.toolbutton import ToolButton
from sugar3.graphics.icon import Icon


_PREVIOUS_PAGE = 0
_NEXT_PAGE = 1


if not hasattr(GObject.ParamFlags, 'READWRITE'):
    GObject.ParamFlags.READWRITE = GObject.ParamFlags.WRITABLE | \
        GObject.ParamFlags.READABLE


class _TrayViewport(Gtk.Viewport):

    __gproperties__ = {
        'scrollable': (bool, None, None, False, GObject.ParamFlags.READABLE),
        'can-scroll-prev': (bool, None, None, False,
                            GObject.ParamFlags.READABLE),
        'can-scroll-next': (bool, None, None, False,
                            GObject.ParamFlags.READABLE),
    }

    def __init__(self, orientation):
        self.orientation = orientation
        self._scrollable = False
        self._can_scroll_next = False
        self._can_scroll_prev = False

        Gtk.Viewport.__init__(self)

        self.set_shadow_type(Gtk.ShadowType.NONE)

        self.traybar = Gtk.Toolbar()
        self.traybar.set_orientation(orientation)
        self.traybar.set_show_arrow(False)
        self.add(self.traybar)
        self.traybar.show()

        self.connect('size-allocate', self._size_allocate_cb)

        if self.orientation == Gtk.Orientation.HORIZONTAL:
            adj = self.get_hadjustment()
        else:
            adj = self.get_vadjustment()
        adj.connect('changed', self._adjustment_changed_cb)
        adj.connect('value-changed', self._adjustment_changed_cb)

    def scroll(self, direction):
        if direction == _PREVIOUS_PAGE:
            self._scroll_previous()
        elif direction == _NEXT_PAGE:
            self._scroll_next()

    def scroll_to_item(self, item):
        """This function scrolls the viewport so that item will be visible."""
        assert item in self.traybar.get_children()

        # Get the allocation, and make sure that it is visible
        allocation = item.get_allocation()
        if self.orientation == Gtk.Orientation.HORIZONTAL:
            adj = self.get_hadjustment()
            start = allocation.x
            stop = allocation.x + allocation.width
        else:
            adj = self.get_vadjustment()
            start = allocation.y
            stop = allocation.y + allocation.height

        if start < adj.get_value():
            adj.set_value(start)
        elif stop > adj.get_value() + adj.get_page_size():
            adj.set_value(stop - adj.get_page_size())

    def _scroll_next(self):
        allocation = self.get_allocation()
        if self.orientation == Gtk.Orientation.HORIZONTAL:
            adj = self.get_hadjustment()
            new_value = adj.get_value() + allocation.width
            adj.set_value(min(new_value, adj.get_upper() - allocation.width))
        else:
            adj = self.get_vadjustment()
            new_value = adj.get_value() + allocation.height
            adj.set_value(min(new_value, adj.get_upper() - allocation.height))

    def _scroll_previous(self):
        allocation = self.get_allocation()
        if self.orientation == Gtk.Orientation.HORIZONTAL:
            adj = self.get_hadjustment()
            new_value = adj.get_value() - allocation.width
            adj.set_value(max(adj.get_lower(), new_value))
        else:
            adj = self.get_vadjustment()
            new_value = adj.get_value() - allocation.height
            adj.set_value(max(adj.get_lower(), new_value))

    def do_get_preferred_width(self):
        if self.orientation == Gtk.Orientation.HORIZONTAL:
            min_width, nat_width = Gtk.Viewport.do_get_preferred_width(self)
            return 0, nat_width
        child_minimum, child_natural = self.get_child().get_preferred_size()
        return child_minimum.width, child_natural.width

    def do_get_preferred_height(self):
        if self.orientation != Gtk.Orientation.HORIZONTAL:
            min_height, nat_height = Gtk.Viewport.do_get_preferred_height(self)
            return 0, nat_height
        child_minimum, child_natural = self.get_child().get_preferred_size()
        return child_minimum.height, child_natural.height

    def do_get_property(self, pspec):
        if pspec.name == 'scrollable':
            return self._scrollable
        elif pspec.name == 'can-scroll-next':
            return self._can_scroll_next
        elif pspec.name == 'can-scroll-prev':
            return self._can_scroll_prev

    def _size_allocate_cb(self, viewport, allocation):
        if allocation.width == 1 and allocation.height == 1:
            # HACK: the first time this callback is called 'width' and
            # 'height' are 1 so we mark the Viewport as scrollable and
            # we show the Prev / Next buttons
            return

        bar_minimum, bar_natural = self.traybar.get_preferred_size()

        if self.orientation == Gtk.Orientation.HORIZONTAL:
            scrollable = bar_minimum.width > allocation.width
        else:
            scrollable = bar_minimum.height > allocation.height

        if scrollable != self._scrollable:
            self._scrollable = scrollable
            self.notify('scrollable')

    def _adjustment_changed_cb(self, adjustment):
        if adjustment.get_value() <= adjustment.get_lower():
            can_scroll_prev = False
        else:
            can_scroll_prev = True

        if adjustment.get_value() + adjustment.get_page_size() >= \
           adjustment.get_upper():
            can_scroll_next = False
        else:
            can_scroll_next = True

        if can_scroll_prev != self._can_scroll_prev:
            self._can_scroll_prev = can_scroll_prev
            self.notify('can-scroll-prev')

        if can_scroll_next != self._can_scroll_next:
            self._can_scroll_next = can_scroll_next
            self.notify('can-scroll-next')


class _TrayScrollButton(ToolButton):

    __gtype_name__ = 'SugarTrayScrollButton'

    def __init__(self, icon_name, scroll_direction):
        ToolButton.__init__(self)
        self._viewport = None

        self._scroll_direction = scroll_direction

        self.set_size_request(style.GRID_CELL_SIZE, style.GRID_CELL_SIZE)

        self.icon = Icon(icon_name=icon_name,
                         pixel_size=style.SMALL_ICON_SIZE)
        # The alignment is a hack to work around Gtk.ToolButton code
        # that sets the icon_size when the icon_widget is a Gtk.Image
        alignment = Gtk.Alignment(xalign=0.5, yalign=0.5)
        alignment.add(self.icon)
        self.set_icon_widget(alignment)
        alignment.show_all()

        self.connect('clicked', self._clicked_cb)

    def set_viewport(self, viewport):
        self._viewport = viewport
        self._viewport.connect('notify::scrollable',
                               self._viewport_scrollable_changed_cb)

        if self._scroll_direction == _PREVIOUS_PAGE:
            self._viewport.connect('notify::can-scroll-prev',
                                   self._viewport_can_scroll_dir_changed_cb)
            self.set_sensitive(self._viewport.props.can_scroll_prev)
        else:
            self._viewport.connect('notify::can-scroll-next',
                                   self._viewport_can_scroll_dir_changed_cb)
            self.set_sensitive(self._viewport.props.can_scroll_next)

    def _viewport_scrollable_changed_cb(self, viewport, pspec):
        self.props.visible = self._viewport.props.scrollable

    def _viewport_can_scroll_dir_changed_cb(self, viewport, pspec):
        if self._scroll_direction == _PREVIOUS_PAGE:
            sensitive = self._viewport.props.can_scroll_prev
        else:
            sensitive = self._viewport.props.can_scroll_next

        self.set_sensitive(sensitive)

    def _clicked_cb(self, button):
        self._viewport.scroll(self._scroll_direction)

    viewport = property(fset=set_viewport)


ALIGN_TO_START = 0
ALIGN_TO_END = 1


class HTray(Gtk.EventBox):

    __gtype_name__ = 'SugarHTray'

    __gproperties__ = {
        'align': (int, None, None, 0, 1, ALIGN_TO_START,
                  GObject.ParamFlags.READWRITE |
                  GObject.ParamFlags.CONSTRUCT_ONLY),
        'drag-active': (bool, None, None, False, GObject.ParamFlags.READWRITE),
    }

    def __init__(self, **kwargs):
        self._drag_active = False
        self.align = ALIGN_TO_START

        Gtk.EventBox.__init__(self, **kwargs)

        self._box = Gtk.HBox()
        self.add(self._box)
        self._box.show()

        scroll_left = _TrayScrollButton('go-left', _PREVIOUS_PAGE)
        self._box.pack_start(scroll_left, False, False, 0)

        self._viewport = _TrayViewport(Gtk.Orientation.HORIZONTAL)
        self._box.pack_start(self._viewport, True, True, 0)
        self._viewport.show()

        scroll_right = _TrayScrollButton('go-right', _NEXT_PAGE)
        self._box.pack_start(scroll_right, False, False, 0)

        scroll_left.viewport = self._viewport
        scroll_right.viewport = self._viewport

        if self.align == ALIGN_TO_END:
            spacer = Gtk.SeparatorToolItem()
            spacer.set_size_request(0, 0)
            spacer.props.draw = False
            spacer.set_expand(True)
            self._viewport.traybar.insert(spacer, 0)
            spacer.show()

    def do_set_property(self, pspec, value):
        if pspec.name == 'align':
            self.align = value
        elif pspec.name == 'drag-active':
            self._set_drag_active(value)
        else:
            raise AssertionError

    def do_get_property(self, pspec):
        if pspec.name == 'align':
            return self.align
        elif pspec.name == 'drag-active':
            return self._drag_active
        else:
            raise AssertionError

    def _set_drag_active(self, active):
        if self._drag_active != active:
            self._drag_active = active
            if self._drag_active:
                self._viewport.traybar.modify_bg(
                    Gtk.StateType.NORMAL,
                    style.COLOR_BLACK.get_gdk_color())
            else:
                self._viewport.traybar.modify_bg(Gtk.StateType.NORMAL, None)

    def get_children(self):
        children = self._viewport.traybar.get_children()[:]
        if self.align == ALIGN_TO_END:
            children = children[1:]
        return children

    def add_item(self, item, index=-1):
        if self.align == ALIGN_TO_END and index > -1:
            index += 1
        self._viewport.traybar.insert(item, index)

    def remove_item(self, item):
        self._viewport.traybar.remove(item)

    def get_item_index(self, item):
        index = self._viewport.traybar.get_item_index(item)
        if self.align == ALIGN_TO_END:
            index -= 1
        return index

    def scroll_to_item(self, item):
        self._viewport.scroll_to_item(item)


if hasattr(HTray, 'set_css_name'):
    HTray.set_css_name('htray')


class VTray(Gtk.EventBox):

    __gtype_name__ = 'SugarVTray'

    __gproperties__ = {
        'align': (int, None, None, 0, 1, ALIGN_TO_START,
                  GObject.ParamFlags.READWRITE |
                  GObject.ParamFlags.CONSTRUCT_ONLY),
        'drag-active': (bool, None, None, False, GObject.ParamFlags.READWRITE),
    }

    def __init__(self, **kwargs):
        self._drag_active = False
        self.align = ALIGN_TO_START

        Gtk.EventBox.__init__(self, **kwargs)

        self._box = Gtk.VBox()
        self.add(self._box)
        self._box.show()

        scroll_up = _TrayScrollButton('go-up', _PREVIOUS_PAGE)
        self._box.pack_start(scroll_up, False, False, 0)

        self._viewport = _TrayViewport(Gtk.Orientation.VERTICAL)
        self._box.pack_start(self._viewport, True, True, 0)
        self._viewport.show()

        scroll_down = _TrayScrollButton('go-down', _NEXT_PAGE)
        self._box.pack_start(scroll_down, False, False, 0)

        scroll_up.viewport = self._viewport
        scroll_down.viewport = self._viewport

        if self.align == ALIGN_TO_END:
            spacer = Gtk.SeparatorToolItem()
            spacer.set_size_request(0, 0)
            spacer.props.draw = False
            spacer.set_expand(True)
            self._viewport.traybar.insert(spacer, 0)
            spacer.show()

    def do_set_property(self, pspec, value):
        if pspec.name == 'align':
            self.align = value
        elif pspec.name == 'drag-active':
            self._set_drag_active(value)
        else:
            raise AssertionError

    def do_get_property(self, pspec):
        if pspec.name == 'align':
            return self.align
        elif pspec.name == 'drag-active':
            return self._drag_active
        else:
            raise AssertionError

    def _set_drag_active(self, active):
        if self._drag_active != active:
            self._drag_active = active
            if self._drag_active:
                self._viewport.traybar.modify_bg(
                    Gtk.StateType.NORMAL,
                    style.COLOR_BLACK.get_gdk_color())
            else:
                self._viewport.traybar.modify_bg(Gtk.StateType.NORMAL, None)

    def get_children(self):
        children = self._viewport.traybar.get_children()[:]
        if self.align == ALIGN_TO_END:
            children = children[1:]
        return children

    def add_item(self, item, index=-1):
        if self.align == ALIGN_TO_END and index > -1:
            index += 1
        self._viewport.traybar.insert(item, index)

    def remove_item(self, item):
        self._viewport.traybar.remove(item)

    def get_item_index(self, item):
        index = self._viewport.traybar.get_item_index(item)
        if self.align == ALIGN_TO_END:
            index -= 1
        return index

    def scroll_to_item(self, item):
        self._viewport.scroll_to_item(item)


if hasattr(VTray, 'set_css_name'):
    VTray.set_css_name('VTray')


class TrayButton(ToolButton):

    def __init__(self, **kwargs):
        ToolButton.__init__(self, **kwargs)


class _IconWidget(Gtk.EventBox):

    __gtype_name__ = 'SugarTrayIconWidget'

    def __init__(self, icon_name=None, xo_color=None):
        Gtk.EventBox.__init__(self)

        self.set_app_paintable(True)
        self.add_events(Gdk.EventMask.BUTTON_PRESS_MASK |
                        Gdk.EventMask.TOUCH_MASK |
                        Gdk.EventMask.BUTTON_RELEASE_MASK)

        self._icon = Icon(icon_name=icon_name, xo_color=xo_color,
                          pixel_size=style.STANDARD_ICON_SIZE)
        self.add(self._icon)
        self._icon.show()

    def do_draw(self, cr):
        palette = self.get_parent().palette

        if palette and palette.is_up():
            allocation = self.get_allocation()
            # draw a black background, has been done by the engine before
            cr.set_source_rgb(0, 0, 0)
            cr.rectangle(0, 0, allocation.width, allocation.height)
            cr.paint()

        Gtk.EventBox.do_draw(self, cr)

        if palette and palette.is_up():
            invoker = palette.props.invoker
            invoker.draw_rectangle(cr, palette)

        return False

    def get_icon(self):
        return self._icon


class TrayIcon(Gtk.ToolItem):

    __gtype_name__ = 'SugarTrayIcon'

    def __init__(self, icon_name=None, xo_color=None):
        Gtk.ToolItem.__init__(self)

        self._icon_widget = _IconWidget(icon_name, xo_color)
        self.add(self._icon_widget)
        self._icon_widget.show()

        self._palette_invoker = ToolInvoker(self)

        self.set_size_request(style.GRID_CELL_SIZE, style.GRID_CELL_SIZE)

        self.connect('destroy', self.__destroy_cb)

    def __destroy_cb(self, icon):
        if self._palette_invoker is not None:
            self._palette_invoker.detach()

    def create_palette(self):
        return None

    def get_palette(self):
        return self._palette_invoker.palette

    def set_palette(self, palette):
        self._palette_invoker.palette = palette

    palette = GObject.Property(
        type=object, setter=set_palette, getter=get_palette)

    def get_palette_invoker(self):
        return self._palette_invoker

    def set_palette_invoker(self, palette_invoker):
        self._palette_invoker.detach()
        self._palette_invoker = palette_invoker

    palette_invoker = GObject.Property(
        type=object, setter=set_palette_invoker, getter=get_palette_invoker)

    def get_icon(self):
        return self._icon_widget.get_icon()
    icon = property(get_icon, None)



# Copyright (C) 2007, Red Hat, Inc.
# Copyright (C) 2008, Benjamin Berg <benjamin@sipsolutions.net>
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

import gettext
from gi.repository import Gdk
from gi.repository import GdkPixbuf
from gi.repository import Gtk
from gi.repository import GObject
import struct
import logging

from sugar3.graphics import style
from sugar3.graphics.icon import Icon
from sugar3.graphics.palette import Palette, ToolInvoker, WidgetInvoker


def _(msg):
    return gettext.dgettext('sugar-toolkit-gtk3', msg)


if not hasattr(GObject.ParamFlags, 'READWRITE'):
    GObject.ParamFlags.READWRITE = GObject.ParamFlags.WRITABLE | \
        GObject.ParamFlags.READABLE


def get_svg_color_string(color):
    return '#%.2X%.2X%.2X' % (color.red // 257, color.green // 257,
                              color.blue // 257)


class _ColorButton(Gtk.Button):
    """
    This is a ColorButton for Sugar. It is similar to the Gtk.ColorButton,
    but does not have any alpha support.
    Instead of a color selector dialog it will pop up a Sugar palette.

    As a preview an sugar3.graphics.Icon is used. The fill color will be set to
    the current color, and the stroke color is set to the font color.
    """

    __gtype_name__ = 'SugarColorButton'
    __gsignals__ = {'color-set': (GObject.SignalFlags.RUN_FIRST, None,
                                  tuple())}

    def __init__(self, **kwargs):
        self._title = _('Choose a color')
        self._color = Gdk.Color(0, 0, 0)
        self._has_palette = True
        self._has_invoker = True
        self._palette = None
        self._accept_drag = True

        self._preview = Icon(icon_name='color-preview',
                             pixel_size=style.STANDARD_ICON_SIZE)

        GObject.GObject.__init__(self, **kwargs)

        # FIXME Drag and drop is not working, SL #3796
        if self._accept_drag:
            self.drag_dest_set(Gtk.DestDefaults.MOTION |
                               Gtk.DestDefaults.HIGHLIGHT |
                               Gtk.DestDefaults.DROP,
                               [Gtk.TargetEntry.new(
                                'application/x-color', 0, 0)],
                               Gdk.DragAction.COPY)
        self.drag_source_set(Gdk.ModifierType.BUTTON1_MASK |
                             Gdk.ModifierType.BUTTON3_MASK,
                             [Gtk.TargetEntry.new(
                                 'application/x-color', 0, 0)],
                             Gdk.DragAction.COPY)
        self.connect('drag-data-received', self.__drag_data_received_cb)
        self.connect('drag-data-get', self.__drag_data_get_cb)

        self._preview.fill_color = get_svg_color_string(self._color)
        self._preview.stroke_color = self._get_fg_style_color_str()
        self.set_image(self._preview)

        if self._has_palette and self._has_invoker:
            self._invoker = WidgetInvoker(self)
            # FIXME: This is a hack.
            self._invoker.has_rectangle_gap = lambda: False
            self._invoker.palette = self._palette

    def create_palette(self):
        '''
        Create a new palette with selected color and title.
        (Here the title is 'Choose a color' and the bgcolor
        is black.)
        '''
        if self._has_palette:
            self._palette = _ColorPalette(color=self._color,
                                          primary_text=self._title)
            self._palette.connect('color-set', self.__palette_color_set_cb)
            self._palette.connect('notify::color', self.
                                  __palette_color_changed)

        return self._palette

    def __palette_color_set_cb(self, palette):
        self.emit('color-set')

    def __palette_color_changed(self, palette, pspec):
        self.color = self._palette.color

    def do_style_set(self, previous_style):
        self._preview.stroke_color = self._get_fg_style_color_str()

    def _get_fg_style_color_str(self):
        context = self.get_style_context()
        fg_color = context.get_color(Gtk.StateType.NORMAL)
        # the color components are stored as float values between 0.0 and 1.0
        return '#%.2X%.2X%.2X' % (int(fg_color.red * 255),
                                  int(fg_color.green * 255),
                                  int(fg_color.blue * 255))

    def set_color(self, color):
        assert isinstance(color, Gdk.Color)

        if self._color.red == color.red and \
           self._color.green == color.green and \
           self._color.blue == color.blue:
            return

        self._color = Gdk.Color(color.red, color.green, color.blue)
        self._preview.fill_color = get_svg_color_string(self._color)
        if self._palette:
            self._palette.props.color = self._color
        self.notify('color')

    def get_color(self):
        return self._color

    color = GObject.Property(type=object, getter=get_color, setter=set_color)

    def set_icon_name(self, icon_name):
        '''
        Sets the icon for the tool button from a named themed icon.
        If it is none then no icon will be shown.

        Args:
            icon_name(string): The name for a themed icon.
            It can be set as 'None' too.

        Example:
            set_icon_name('view-radial')
        '''
        self._preview.props.icon_name = icon_name

    def get_icon_name(self):
        '''
        The get_icon_name() method returns the value of the icon_name
        property that contains the name of a themed icon or None.
        '''
        return self._preview.props.icon_name

    icon_name = GObject.Property(type=str,
                                 getter=get_icon_name, setter=set_icon_name)

    def set_icon_size(self, pixel_size):
        self._preview.props.pixel_size = pixel_size

    def get_icon_size(self):
        return self._preview.props.pixel_size

    icon_size = GObject.Property(type=int,
                                 getter=get_icon_size, setter=set_icon_size)

    def set_title(self, title):
        self._title = title
        if self._palette:
            self._palette.primary_text = self._title

    def get_title(self):
        return self._title

    title = GObject.Property(type=str, getter=get_title, setter=set_title)

    def _set_has_invoker(self, has_invoker):
        self._has_invoker = has_invoker

    def _get_has_invoker(self):
        return self._has_invoker

    has_invoker = GObject.Property(type=bool, default=True,
                                   flags=GObject.ParamFlags.READWRITE |
                                   GObject.ParamFlags.CONSTRUCT_ONLY,
                                   getter=_get_has_invoker,
                                   setter=_set_has_invoker)

    def _set_has_palette(self, has_palette):
        self._has_palette = has_palette

    def _get_has_palette(self):
        return self._has_palette

    has_palette = GObject.Property(type=bool, default=True,
                                   flags=GObject.ParamFlags.READWRITE |
                                   GObject.ParamFlags.CONSTRUCT_ONLY,
                                   getter=_get_has_palette,
                                   setter=_set_has_palette)

    def _set_accept_drag(self, accept_drag):
        self._accept_drag = accept_drag

    def _get_accept_drag(self):
        return self._accept_drag

    accept_drag = GObject.Property(type=bool, default=True,
                                   flags=GObject.ParamFlags.READWRITE |
                                   GObject.ParamFlags.CONSTRUCT_ONLY,
                                   getter=_get_accept_drag,
                                   setter=_set_accept_drag)

    def __drag_begin_cb(self, widget, context):
        # Drag and Drop
        pixbuf = GdkPixbuf.Pixbuf(GdkPixbuf.Colorspace.RGB, True, 8,
                                  style.SMALL_ICON_SIZE,
                                  style.SMALL_ICON_SIZE)

        red = self._color.red / 257
        green = self._color.green / 257
        blue = self._color.blue / 257

        pixbuf.fill(red << 24 + green << 16 + blue << 8 + 0xff)

        context.set_icon_pixbuf(pixbuf)

    def __drag_data_get_cb(self, widget, context, selection_data, info, time):
        data = struct.pack('=HHHH', self._color.red, self._color.green,
                           self._color.blue, 65535)
        selection_data.set(selection_data.target, 16, data)

    def __drag_data_received_cb(self, widget, context, x, y, selection_data,
                                info, time):
        if len(selection_data.data) != 8:
            return

        dropped = selection_data.data
        red = struct.unpack_from('=H', dropped, 0)[0]
        green = struct.unpack_from('=H', dropped, 2)[0]
        blue = struct.unpack_from('=H', dropped, 4)[0]
        # dropped[6] and dropped[7] is alpha, but we ignore the alpha channel

        color = Gdk.Color(red, green, blue)
        self.set_color(color)


class _ColorPalette(Palette):
    """This is a color picker palette. It will usually be used indirectly
    trough a sugar3.graphics.ColorButton.
    """
    _RED = 0
    _GREEN = 1
    _BLUE = 2

    __gtype_name__ = 'SugarColorPalette'

    # The color-set signal is emitted when the user is finished selecting
    # a color.
    __gsignals__ = {'color-set': (GObject.SignalFlags.RUN_FIRST, None,
                                  tuple())}

    def __init__(self, **kwargs):
        self._color = Gdk.Color(0, 0, 0)
        self._previous_color = self._color.copy()
        self._scales = None

        Palette.__init__(self, **kwargs)

        self.connect('popup', self.__popup_cb)
        self.connect('popdown', self.__popdown_cb)

        self._picker_hbox = Gtk.HBox()
        alignment = Gtk.Alignment()
        alignment.set_padding(0, 0, style.DEFAULT_SPACING,
                              style.DEFAULT_SPACING)
        alignment.add(self._picker_hbox)
        self.set_content(alignment)
        alignment.show()

        self._swatch_tray = Gtk.Table()

        self._picker_hbox.pack_start(self._swatch_tray, True, True, 0)
        self._picker_hbox.pack_start(Gtk.VSeparator(), True, True,
                                     padding=style.DEFAULT_SPACING)

        self._chooser_table = Gtk.Table(3, 2)
        self._chooser_table.set_col_spacing(0, style.DEFAULT_PADDING)

        self._scales = []
        self._scales.append(
            self._create_color_scale(_('Red'), self._RED, 0))
        self._scales.append(
            self._create_color_scale(_('Green'), self._GREEN, 1))
        self._scales.append(
            self._create_color_scale(_('Blue'), self._BLUE, 2))

        self._picker_hbox.add(self._chooser_table)

        self._picker_hbox.show_all()

        self._build_swatches()

    def _create_color_scale(self, text, color, row):
        label = Gtk.Label(label=text)
        label.props.xalign = 1.0
        scale = Gtk.HScale()
        scale.set_size_request(style.zoom(250), -1)
        scale.set_draw_value(False)
        scale.set_range(0, 1.0)
        scale.set_increments(0.1, 0.2)

        if color == self._RED:
            scale.set_value(self._color.red / 65535.0)
        elif color == self._GREEN:
            scale.set_value(self._color.green / 65535.0)
        elif color == self._BLUE:
            scale.set_value(self._color.blue / 65535.0)

        scale.connect('value-changed',
                      self.__scale_value_changed_cb,
                      color)
        self._chooser_table.attach(label, 0, 1, row, row + 1)
        self._chooser_table.attach(scale, 1, 2, row, row + 1)

        return scale

    def _build_swatches(self):
        for child in self._swatch_tray.get_children():
            child.destroy()

        # Use a hardcoded list of colors for now.
        colors = ['#ed2529', '#69bc47', '#3c54a3',
                  '#f57f25', '#0b6b3a', '#00a0c6',
                  '#f6eb1a', '#b93f94', '#5b4a9c',
                  '#000000', '#919496', '#ffffff']

        # We want 3 rows of colors.
        rows = 3
        i = 0
        self._swatch_tray.props.n_rows = rows
        self._swatch_tray.props.n_columns = (len(colors) + rows - 1) / rows
        for color in colors:
            button = _ColorButton(has_palette=False,
                                  color=Gdk.color_parse(color),
                                  accept_drag=False,
                                  icon_size=style.STANDARD_ICON_SIZE)
            button.set_relief(Gtk.ReliefStyle.NONE)
            self._swatch_tray.attach(button,
                                     i % rows, i % rows + 1,
                                     i / rows, i / rows + 1,
                                     yoptions=0, xoptions=0)
            button.connect('clicked', self.__swatch_button_clicked_cb)
            i += 1

        self._swatch_tray.show_all()

    def __popup_cb(self, palette):
        self._previous_color = self._color.copy()

    def __popdown_cb(self, palette):
        self.emit('color-set')

    def __scale_value_changed_cb(self, widget, color):
        new_color = self._color.copy()
        if color == self._RED:
            new_color.red = int(65535 * widget.get_value())
        elif color == self._GREEN:
            new_color.green = int(65535 * widget.get_value())
        elif color == self._BLUE:
            new_color.blue = int(65535 * widget.get_value())
        self.color = new_color

    def do_key_press_event(self, event):
        if event.keyval == Gdk.KEY_Escape:
            self.props.color = self._previous_color
            self.popdown(immediate=True)
            return True
        elif event.keyval == Gdk.KEY_Return:
            self.popdown(immediate=True)
            return True
        return False

    def __swatch_button_clicked_cb(self, button):
        self.props.color = button.get_color()

    def set_color(self, color):
        assert isinstance(color, Gdk.Color)

        if self._color.red == color.red and \
           self._color.green == color.green and \
           self._color.blue == color.blue:
            return

        self._color = color.copy()

        if self._scales:
            self._scales[self._RED].set_value(self._color.red / 65535.0)
            self._scales[self._GREEN].set_value(self._color.green / 65535.0)
            self._scales[self._BLUE].set_value(self._color.blue / 65535.0)

        self.notify('color')

    def get_color(self):
        return self._color

    color = GObject.Property(type=object, getter=get_color, setter=set_color)


def _add_accelerator(tool_button):
    if not tool_button.props.accelerator or not tool_button.get_toplevel() or \
            not tool_button.get_child():
        return

    # TODO: should we remove the accelerator from the prev top level?

    if not hasattr(tool_button.get_toplevel(), 'sugar_accel_group'):
        logging.debug('No Gtk.AccelGroup in the top level window.')
        return

    accel_group = tool_button.get_toplevel().sugar_accel_group
    keyval, mask = Gtk.accelerator_parse(tool_button.props.accelerator)
    # the accelerator needs to be set at the child, so the Gtk.AccelLabel
    # in the palette can pick it up.
    tool_button.get_child(
    ).add_accelerator('clicked', accel_group, keyval, mask,
                      Gtk.AccelFlags.LOCKED | Gtk.AccelFlags.VISIBLE)


def _hierarchy_changed_cb(tool_button, previous_toplevel):
    _add_accelerator(tool_button)


def setup_accelerator(tool_button):
    _add_accelerator(tool_button)
    tool_button.connect('hierarchy-changed', _hierarchy_changed_cb)


class ColorToolButton(Gtk.ToolItem):
    # This not ideal. It would be better to subclass Gtk.ToolButton, however
    # the python bindings do not seem to be powerfull enough for that.
    # (As we need to change a variable in the class structure.)

    __gtype_name__ = 'SugarColorToolButton'
    __gsignals__ = {'color-set': (GObject.SignalFlags.RUN_FIRST, None,
                                  tuple())}

    def __init__(self, icon_name='color-preview', **kwargs):
        self._accelerator = None
        self._tooltip = None
        self._palette_invoker = ToolInvoker()
        self._palette = None

        GObject.GObject.__init__(self, **kwargs)

        # The Gtk.ToolButton has already added a normal button.
        # Replace it with a ColorButton
        color_button = _ColorButton(icon_name=icon_name, has_invoker=False)
        self.add(color_button)
        color_button.show()

        # The following is so that the behaviour on the toolbar is correct.
        color_button.set_relief(Gtk.ReliefStyle.NONE)
        color_button.icon_size = style.STANDARD_ICON_SIZE

        self._palette_invoker.attach_tool(self)
        self._palette_invoker.props.toggle_palette = True
        self._palette_invoker.props.lock_palette = True

        # This widget just proxies the following properties to the colorbutton
        color_button.connect('notify::color', self.__notify_change)
        color_button.connect('notify::icon-name', self.__notify_change)
        color_button.connect('notify::icon-size', self.__notify_change)
        color_button.connect('notify::title', self.__notify_change)
        color_button.connect('color-set', self.__color_set_cb)
        color_button.connect('can-activate-accel',
                             self.__button_can_activate_accel_cb)

    def __button_can_activate_accel_cb(self, button, signal_id):
        # Accept activation via accelerators regardless of this widget's state
        return True

    def set_accelerator(self, accelerator):
        '''
        Sets keyboard shortcut that activates this button.

        Args:
            accelerator(string): accelerator to be set. Should be in
            form <modifier>Letter
            Find about format here :
            https://developer.gnome.org/gtk3/stable/gtk3-Keyboard-Accelerators.html#gtk-accelerator-parse

        Example:
            set_accelerator(self, 'accel')
        '''
        self._accelerator = accelerator
        setup_accelerator(self)

    def get_accelerator(self):
        '''
        Returns the above accelerator string.
        '''
        return self._accelerator

    accelerator = GObject.Property(type=str, setter=set_accelerator,
                                   getter=get_accelerator)

    def create_palette(self):
        '''
        The create_palette function is called when the palette needs to be
        invoked.  For example, when the user has right clicked the icon or
        the user has hovered over the icon for a long time.

        The create_palette will only be called once or zero times.  The palette
        returned will be stored and re-used if the user invokes the palette
        multiple times.

        Your create_palette implementation does not need to
        :any:`Gtk.Widget.show` the palette, as this will be done by the
        invoker.  However, you still need to show
        the menu items, etc that you place in the palette.

        Returns:

            sugar3.graphics.palette.Palette, or None to indicate that you
            do not want a palette shown

        The default implementation returns None, to indicate no palette should
        be shown.
        '''
        self._palette = self.get_child().create_palette()
        return self._palette

    def get_palette_invoker(self):
        return self._palette_invoker

    def set_palette_invoker(self, palette_invoker):
        self._palette_invoker.detach()
        self._palette_invoker = palette_invoker

    palette_invoker = GObject.Property(
        type=object, setter=set_palette_invoker, getter=get_palette_invoker)

    def set_expanded(self, expanded):
        box = self.toolbar_box
        if not box:
            return

        if not expanded:
            self._palette_invoker.notify_popdown()
            return

        if box.expanded_button is not None:
            box.expanded_button.queue_draw()
            if box.expanded_button != self:
                box.expanded_button.set_expanded(False)
        box.expanded_button = self

    def get_toolbar_box(self):
        parent = self.get_parent()
        if not hasattr(parent, 'owner'):
            return None
        return parent.owner

    toolbar_box = property(get_toolbar_box)

    def set_color(self, color):
        '''
        Sets the color of the colorbutton
        '''
        self.get_child().props.color = color

    def get_color(self):
        '''
        Gets the above set color string.
        '''
        return self.get_child().props.color

    color = GObject.Property(type=object, getter=get_color, setter=set_color)

    def set_icon_name(self, icon_name):
        '''
        Sets the icon for the tool button from a named themed icon.
        If it is none then no icon will be shown.

        Args:
            icon_name(string): The name for a themed icon.
            It can be set as 'None' too.

        Example:
            set_icon_name('view-radial')
        '''
        self.get_child().props.icon_name = icon_name

    def get_icon_name(self):
        '''
        The get_icon_name() method returns the value of the icon_name
        property that contains the name of a themed icon or None.
        '''
        return self.get_child().props.icon_name

    icon_name = GObject.Property(type=str,
                                 getter=get_icon_name, setter=set_icon_name)

    def set_icon_size(self, icon_size):
        '''
        Sets the size of icons in the colorbutton.
        '''
        self.get_child().props.icon_size = icon_size

    def get_icon_size(self):
        '''
        Gets the size of icons in the colorbutton.
        '''
        return self.get_child().props.icon_size

    icon_size = GObject.Property(type=int,
                                 getter=get_icon_size, setter=set_icon_size)

    def set_title(self, title):
        '''
        The set_title() method sets the "title" property to the value of
        title. The "title" property contains the string that is used to
        set the colorbutton title.
        '''
        self.get_child().props.title = title

    def get_title(self):
        '''
        Return the above title string.
        '''
        return self.get_child().props.title

    title = GObject.Property(type=str, getter=get_title, setter=set_title)

    def do_draw(self, cr):
        if self._palette and self._palette.is_up():
            allocation = self.get_allocation()
            # draw a black background, has been done by the engine before
            cr.set_source_rgb(0, 0, 0)
            cr.rectangle(0, 0, allocation.width, allocation.height)
            cr.paint()

        Gtk.ToolItem.do_draw(self, cr)

        if self._palette and self._palette.is_up():
            invoker = self._palette.props.invoker
            invoker.draw_rectangle(cr, self._palette)

        return False

    def __notify_change(self, widget, pspec):
        self.notify(pspec.name)

    def __color_set_cb(self, widget):
        self.emit('color-set')



# Copyright (C) 2006-2007 Red Hat, Inc.
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""
UNSTABLE. It should really be internal to the Activity class.
"""

import logging

import dbus
import dbus.service


_ACTIVITY_SERVICE_NAME = 'org.laptop.Activity'
_ACTIVITY_SERVICE_PATH = '/org/laptop/Activity'
_ACTIVITY_INTERFACE = 'org.laptop.Activity'


class ActivityService(dbus.service.Object):
    """Base dbus service object that each Activity uses to export dbus methods.

    The dbus service is separate from the actual Activity object so that we can
    tightly control what stuff passes through the dbus python bindings."""

    def __init__(self, activity):
        """Initialise the service for the given activity

        activity -- sugar3.activity.activity.Activity instance

        Creates dbus services that use the instance's activity_id
        as discriminants among all active services
        of this type.  That is, the services are all available
        as names/paths derived from the instance's activity_id.

        The various methods exposed on dbus are just forwarded
        to the client Activity object's equally-named methods.
        """
        activity.realize()

        activity_id = activity.get_id()
        service_name = _ACTIVITY_SERVICE_NAME + activity_id
        object_path = _ACTIVITY_SERVICE_PATH + '/' + activity_id

        bus = dbus.SessionBus()
        bus_name = dbus.service.BusName(service_name, bus=bus)
        dbus.service.Object.__init__(self, bus_name, object_path)

        self._activity = activity

    @dbus.service.method(_ACTIVITY_INTERFACE)
    def SetActive(self, active):
        logging.debug('ActivityService.set_active: %s.', active)
        self._activity.props.active = active

    @dbus.service.method(_ACTIVITY_INTERFACE)
    def InviteContact(self, account_path, contact_id):
        self._activity.invite(account_path, contact_id)

    @dbus.service.method(_ACTIVITY_INTERFACE)
    def HandleViewSource(self):
        self._activity.handle_view_source()

    @dbus.service.method(_ACTIVITY_INTERFACE,
                         async_callbacks=('async_cb', 'async_err_cb'))
    def GetDocumentPath(self, async_cb, async_err_cb):
        try:
            self._activity.get_document_path(async_cb, async_err_cb)
        except Exception as e:
            async_err_cb(e)



# Copyright (C) 2007, Red Hat, Inc.
# Copyright (C) 2009 Aleksey Lim
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""Sugar content bundles

UNSTABLE.
"""

from six.moves import urllib
from six.moves.configparser import ConfigParser

import tempfile
import os

from sugar3 import env
from sugar3.bundle.bundle import Bundle, MalformedBundleException

from sugar3.bundle.bundleversion import NormalizedVersion
from sugar3.bundle.bundleversion import InvalidVersionError


class ContentBundle(Bundle):
    """A Sugar content bundle

    See http://wiki.laptop.org/go/Content_bundles for details
    """

    MIME_TYPE = 'application/vnd.olpc-content'

    _zipped_extension = '.xol'
    _unzipped_extension = None
    _infodir = 'library'

    def __init__(self, path):
        Bundle.__init__(self, path)

        self._locale = None
        self._name = None
        self._icon = None
        self._library_version = '0'
        self._activity_start = 'index.html'
        self._global_name = None

        info_file = self.get_file('library/library.info')
        if info_file is None:
            raise MalformedBundleException('No library.info file')
        self._parse_info(info_file)

        if self.get_file(self._activity_start) is None:
            raise MalformedBundleException(
                'Content bundle %s does not have start page %s' %
                (self._path, self._activity_start))

    def _parse_info(self, info_file):
        cp = ConfigParser()
        cp.readfp(info_file)

        section = 'Library'

        if cp.has_option(section, 'name'):
            self._name = cp.get(section, 'name')
        else:
            raise MalformedBundleException(
                'Content bundle %s does not specify a name' % self._path)

        if cp.has_option(section, 'library_version'):
            version = cp.get(section, 'library_version')
            try:
                NormalizedVersion(version)
            except InvalidVersionError:
                raise MalformedBundleException(
                    'Content bundle %s has invalid version number %s' %
                    (self._path, version))
            self._library_version = version

        if cp.has_option(section, 'locale'):
            self._locale = cp.get(section, 'locale')

        if cp.has_option(section, 'global_name'):
            self._global_name = cp.get(section, 'global_name')

        if cp.has_option(section, 'icon'):
            self._icon = cp.get(section, 'icon')

        # Compatibility with old content bundles
        if self._global_name is not None \
                and cp.has_option(section, 'bundle_class'):
            self._global_name = cp.get(section, 'bundle_class')

        if cp.has_option(section, 'activity_start'):
            self._activity_start = cp.get(section, 'activity_start')

        if self._global_name is None:
            raise MalformedBundleException(
                'Content bundle %s must specify global_name' % self._path)

    def get_name(self):
        return self._name

    def get_library_version(self):
        return self._library_version

    def get_locale(self):
        return self._locale

    def get_activity_start(self):
        return self._activity_start

    def get_icon(self):
        if not self._icon:
            return None

        icon_path = os.path.join('library', self._icon)
        ext = os.path.splitext(icon_path)[1]
        if ext == '':
            ext = '.svg'
            icon_path += ext

        if self._zip_file is None:
            return os.path.join(self._path, icon_path)
        else:
            icon_data = self.get_file(icon_path).read()
            temp_file, temp_file_path = tempfile.mkstemp(prefix=self._icon,
                                                         suffix=ext)
            os.write(temp_file, icon_data)
            os.close(temp_file)
            return temp_file_path

    def get_start_uri(self):
        path = os.path.join(self.get_path(), self._activity_start)
        return 'file://' + urllib.request.pathname2url(path)

    def get_bundle_id(self):
        return self._global_name

    def get_activity_version(self):
        return self._library_version

    def get_tags(self):
        return None

    def install(self):
        install_path = env.get_user_library_path()
        self._unzip(install_path)
        return os.path.join(install_path, self._zip_root_dir)

    def uninstall(self, force=False, delete_profile=False):
        install_dir = self._path
        self._uninstall(install_dir)

    def is_user_activity(self):
        # All content bundles are installed in user storage
        return True



# Copyright (C) 2007, One Laptop Per Child
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the
# Free Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

'''
The combobox module provides a combo box; a button like widget which
creates a list popup when clicked.  It's best used outside of a
toolbar when the user needs to choose from a *short* list of items.

Example:

.. literalinclude:: ../examples/combobox.py
'''

from gi.repository import GObject
from gi.repository import Gtk
from gi.repository import GdkPixbuf


class ComboBox(Gtk.ComboBox):
    '''
    This class provides a simple wrapper based on the :class:`Gtk.ComboBox`.
    This lets you make a list of items, with a value, label and optionally an
    icon.
    '''
    __gtype_name__ = 'SugarComboBox'

    def __init__(self):
        GObject.GObject.__init__(self)

        self._text_renderer = None
        self._icon_renderer = None

        self._model = Gtk.ListStore(GObject.TYPE_PYOBJECT,
                                    GObject.TYPE_STRING,
                                    GdkPixbuf.Pixbuf,
                                    GObject.TYPE_BOOLEAN)
        self.set_model(self._model)

        self.set_row_separator_func(self._is_separator, None)

    def get_value(self):
        '''
        The value of the currently selected item; the same as the `value`
        argument that was passed to the `append_item` func.

        Returns:
            object, value of selected item
        '''
        row = self.get_active_item()
        if not row:
            return None
        return row[0]

    value = GObject.Property(
        type=object, getter=get_value, setter=None)

    def _get_real_name_from_theme(self, name, size):
        icon_theme = Gtk.IconTheme.get_default()
        valid_, width, height = Gtk.icon_size_lookup(size)
        info = icon_theme.lookup_icon(name, max(width, height), 0)
        if not info:
            raise ValueError('Icon %r not found.' % name)
        fname = info.get_filename()
        del info
        return fname

    def append_item(self, value, text, icon_name=None, file_name=None):
        '''
        This function adds another item to the bottom of the combo box list.

        If either `icon_name` or `file_name` are supplied and icon column
        will be added to the combo box list.

        Args:
            value (object):  the value that will be returned by `get_value`,
                when this item is selected
            text (str):  the user visible label for the item
            icon_name (str):  the name of the icon in the theme to use for
                this item, optional and conflicting with `file_name`
            file_name (str):  the path to a sugar (svg) icon to use for this
                item, optional and conflicting with `icon_name`
        '''
        if not self._icon_renderer and (icon_name or file_name):
            self._icon_renderer = Gtk.CellRendererPixbuf()

            settings = self.get_settings()
            valid_, w, h = Gtk.icon_size_lookup_for_settings(
                settings, Gtk.IconSize.MENU)
            self._icon_renderer.props.stock_size = max(w, h)

            self.pack_start(self._icon_renderer, False)
            self.add_attribute(self._icon_renderer, 'pixbuf', 2)

        if not self._text_renderer and text:
            self._text_renderer = Gtk.CellRendererText()
            self.pack_end(self._text_renderer, True)
            self.add_attribute(self._text_renderer, 'text', 1)

        if icon_name or file_name:
            if text:
                size = Gtk.IconSize.MENU
            else:
                size = Gtk.IconSize.LARGE_TOOLBAR
            valid_, width, height = Gtk.icon_size_lookup(size)

            if icon_name:
                file_name = self._get_real_name_from_theme(icon_name, size)

            pixbuf = GdkPixbuf.Pixbuf.new_from_file_at_size(
                file_name, width, height)
        else:
            pixbuf = None

        self._model.append([value, text, pixbuf, False])

    def append_separator(self):
        '''
        Add a separator to the bottom of the combo box list.  The separator
        can not be selected.
        '''
        self._model.append([0, None, None, True])

    def get_active_item(self):
        '''
        Get the row of data for the currently selected item.

        Returns:
            row of data in the format::

                [value, text, pixbuf, is_separator]
        '''
        index = self.get_active()
        if index == -1:
            index = 0

        row = self._model.iter_nth_child(None, index)
        if not row:
            return None
        return self._model[row]

    def remove_all(self):
        '''
        Remove all list items from the combo box.
        '''
        self._model.clear()

    def _is_separator(self, model, row, data):
        return model[row][3]
